#!/usr/bin/env python3
import numpy as np

import tf
import os
from multiprocessing import Lock
import cv2
import numpy as np
import rospy
from std_msgs.msg import Float32
from sensor_msgs.msg import JointState
from nav_msgs.msg import Odometry, OccupancyGrid
from measurement_pipeline.ros_4_hri_interface import ROS4HRIInterface
from measurement_pipeline.srv import measurement, measurementResponse
from geometry_msgs.msg import Twist, Vector3

# roslaunch drl_navigation drl_navigation.launch
# rosservice call /measurement_server "start: false"
class MeasurementGenerator:
    def __init__(self):
        os.environ['XLA_PYTHON_CLIENT_PREALLOCATE'] = 'false'
        robot_config_name = rospy.get_param('~robot_config_name', 'None')
        robot_config_path = None if robot_config_name == 'None' else robot_config_name
        self.robot_config_path = robot_config_path
        self.max_humans_world = rospy.get_param('max_humans_world', 20)
        self.max_groups_world = rospy.get_param('max_groups_world', 10)
        self.namespace = rospy.get_param('namespace', 'behavior_generator')
        self.namespace_slam = rospy.get_param('~namespace_slam', '/rtabmap')
        self.map_frame = rospy.get_param('~map_frame', 'map')
        self.robot_frame = rospy.get_param('~robot_frame', 'base_footprint')
        self.global_occupancy_map_topic = rospy.get_param('~global_occupancy_map_topic', '/slam/global_occupancy_map')
        self.local_occupancy_map_topic = rospy.get_param('~local_occupancy_map_topic', '/slam/local_occupancy_map')
        self.global_occupancy_map_rl_topic = rospy.get_param('~global_map_rl_navigation_topic', '/slam/global_map_rl_navigation')
        self.local_occupancy_map_rl_topic = rospy.get_param('~local_map_rl_navigation_topic', '/slam/local_map_rl_navigation')
        self.local_social_map_topic = rospy.get_param('~local_robot_social_cost_map_topic', '/slam/local_robot_social_cost_map')
        self.final_goal_frame = rospy.get_param('~final_goal_frame', 'final_goal')
        self.intermediate_waypoint_goal_frame = rospy.get_param('~intermediate_waypoint_goal_frame', 'intermediate_waypoint_goal')
        self.cmd_vel_topic = rospy.get_param('~nav_vel', '/mobile_base_controller/cmd_vel')
        self.time_inference_mpc_topic = rospy.get_param('~time_inference_mpc_topic', '/time_inference_mpc')
        self.time_inference_rl_topic = rospy.get_param('~time_inference_rl_topic', '/time_inference_rl')

        # Odometry
        self.odometry_topic = rospy.get_param('~odometry_topic', '/slam/odometry')
        self.namespace_slam = rospy.get_param('~namespace_slam', '/rtabmap')
        self.diag_timer_rate = rospy.get_param('~diag_timer_rate', 10)
        self.start_experiment_id = rospy.get_param('~start_from_experiment_id', 1)
        self.campaign_name = rospy.get_param('~campaign_name', 'campaign')


        self.joint_states_data = None

        self.go_to_position_data = None
        self.rtabmap_ready = False

        # Maps
        self.local_map_static_list = list()
        self.local_map_static = None
        self.local_map_static_rl_list = list()
        self.local_map_static_rl = None
        
        self.local_map_data = None
        
        self.global_map_data = None

        self.local_social_map = None

        # To get indexes of the joints
        self.get_indexes = False

        # Position and velocity measurement
        self.robot_x_position = 0.
        self.robot_y_position = 0.
        self.robot_z_position = 0.
        self.robot_x_linear_velocity = 0.
        self.robot_y_linear_velocity = 0.
        self.robot_z_linear_velocity = 0.
        self.robot_x_angular_velocity = 0.
        self.robot_y_angular_velocity = 0.
        self.robot_z_angular_velocity = 0.
        self.robot_w_angular_velocity = 0.
        self.robot_yaw_robot_frame = 0.
        self.robot_position_robot_frame = 0.

        # Position/Velocity from odometry
        self.robot_position_list = list()
        self.robot_yaw_position_list = list()
        self.robot_linear_velocity_list = list()
        self.robot_yaw_velocity_list = list()

        # Velocity commands
        self.command_linear_velocity_list = list()
        self.command_yaw_velocity_list = list()
        self.cmd_lin_vel, self.cmd_ang_vel = 0,0
        self.time_inference_rl = 0
        self.time_inference_rl_list = list()
        self.time_inference_mpc = 0
        self.time_inference_mpc_list = list()
        
        # Intermediate waypoint
        self.intermediate_waypoint_list = list()
        self.x_wp = 0.
        self.y_wp = 0.
        self.yaw_wp = 0.

        # Final Goal
        self.final_goal_list = list()
        self.x_final_goal = 0.
        self.y_final_goal = 0.
        self.yaw_final_goal = 0.

        # Position from robot frame
        self.robot_yaw_robot_frame_list = list()
        self.robot_position_robot_frame_list = list()
        
        # Time step measurements
        self.ros_step_list = list()
        self.ros_time_list = list()
        self.robot_time_list = list()

        # Human informations
        self.ros_4_hri_interface = ROS4HRIInterface()
        self.hri_data = None
        self.human_id_default = -1
        self.group_id_default = -1

        self.humans_id = np.ones(self.max_humans_world) * self.human_id_default
        self.humans_group_id = np.ones(self.max_humans_world) * self.group_id_default
        self.humans_group_id_list = list()
        self.groups_id = np.ones(self.max_groups_world) * self.group_id_default
        self.groups_id_list = list()
        self.humans_visible = np.ones(self.max_humans_world)
        self.humans_visible_list = list()
        self.humans_pose = np.zeros((self.max_humans_world, 3))  # x, y, angle
        self.humans_pose_list = list()
        self.groups_pose_list = list()
        self.robot_pose = np.zeros(3)  # x, y, angle
        self.humans_velocity = np.zeros((self.max_humans_world, 2))  # linear, angular
        self.groups_pose = np.zeros((self.max_groups_world, 2))  # x, y
        self.groups_visible = np.ones(self.max_groups_world)
         
        # Social cost of robot
        self.robot_social_cost_list = list()
        self.robot_social_cost = 0.

        self.start_measurement_flag = False
        self.stop_measurement_flag = False 

        self.robot_config = None
        
        self.global_map_static = None
        self.global_map_static_rl = None
        self.global_map_size = 0

        self.action_idx = None
        self.actions = None
        self.controller_status_msg = None

        self.timestamp_tracking= None
        self.timestamp_tracking_latest = None
        self.lock_data_entrance = Lock()
        self.lock_action_publishing = Lock()

        self.init_ros_subscriber_and_publicher()

        # Service
        self.measurement_server = rospy.Service('measurement_server', measurement, self.handle_measurements)

        self.ros_step = 0

        self.tf_broadcaster = tf.TransformBroadcaster() # Publish Transform to look for human
        self.tf_listener = tf.TransformListener() # Listen Transform to look for human

        self._check_all_sensors_ready()
        
        self.actions = []

        self.time_start = None
        # Check what repetition already exist and set self.repetition_counter (ex : if there are already 5 repetition, set self.repetition_counter = 6)
        self.path = self.create_experiment_folder_path(init_path='/home/ros/robot_behavior_ws/experiments/'+self.campaign_name+'/experiments/', counter_exp=self.start_experiment_id)
        self.repetition_counter = 1
        self.flag_repetition_folder_okay = False
        rospy.loginfo("BehaviorGenerator Initialization Ended")


    def init_ros_subscriber_and_publicher(self):
        r''' Initialize the subscribers and publishers '''
        self._subscribers = []
        self._publishers = []
        self._timers = []

        ### ROS Subscribers
        self._subscribers.append(rospy.Subscriber(self.global_occupancy_map_topic, OccupancyGrid, callback=self._global_map_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.local_occupancy_map_topic, OccupancyGrid, callback=self._local_map_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.global_occupancy_map_rl_topic, OccupancyGrid, callback=self._global_map_rl_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.local_occupancy_map_rl_topic, OccupancyGrid, callback=self._local_map_rl_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.odometry_topic, Odometry, callback=self._odometry_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.local_social_map_topic, OccupancyGrid, callback=self._local_social_map_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.cmd_vel_topic, Twist, callback=self._callback_vel_cmd, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.time_inference_rl_topic, Float32, callback=self._callback_time_inference_rl, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.time_inference_mpc_topic, Float32, callback=self._callback_time_inference_mpc, queue_size=1))

        self.step_rate = rospy.Rate(10)

        rospy.loginfo("Initializing the BehaviorGenerator")
 


    def run(self):
        r''' Runs the ros wrapper '''
        init_t = rospy.Time.now()
        while not rospy.is_shutdown():
            self.step(init_t)


    def step(self, init_time):
        r''' Step the ros wrapper'''
        # print('ROS STEP : {}'.format(self.ros_step))

        loop_t = rospy.Time.now()
        if self.time_start is None :
            self.time_start = self.timestamp_odometry_cb
        
        # Set repetition folder
        

        if not self.start_measurement_flag and not self.stop_measurement_flag:
            print('Measurements in standby')
        

        if self.start_measurement_flag :
            if self.path is None :
                print('The path was not defined so the experiment will be saved in the default folder')

            if not self.flag_repetition_folder_okay :
                self.set_repetition_folder_path()
                self.flag_repetition_folder_okay=True

            # Get pose
            self.get_robot_pose()
            self.get_waypoint_pose()
            self.get_final_goal_pose()

            # Storing measurement
            self.store_measurements()
            if self.ros_step % 10 == 0:
                self.log_measurement()

        if self.stop_measurement_flag :
            print('Measurements Stoped')
            self.log_measurement()
            self.stop_measurement_flag = False
            self.flag_repetition_folder_okay=False
            self.reset_logs()

        time_now = rospy.Time.now()
        # rospy.logdebug('BehaviorGenerator time : {}, BehaviorGenerator loop time : {}'.format((time_now - init_time).to_sec(), (time_now - loop_t).to_sec()))
        # rospy.logdebug('Step time : {}'.format((time_now - loop_t).to_sec()))

        self.step_rate.sleep()
        self.ros_step += 1


    def set_repetition_folder_path(self):

        # go through repetition folders from rep 1 upwards until we find an empty one
        self.repetition_counter = 1
        path = self.path + 'repetition_{:06d}'.format(self.repetition_counter) + '/data/'
        while os.path.exists(path):
            self.repetition_counter += 1
            path = self.path + 'repetition_{:06d}'.format(self.repetition_counter) + '/data/'
        os.makedirs(path)

        print('The following repetition will have the given id : {}'.format(path[-2]))


    def create_experiment_folder_path(self, init_path, counter_exp):
        path = init_path + 'experiment_{:06d}'.format(counter_exp) + '/'
        if not os.path.exists(path):
            os.makedirs(path)
        else :
            pass 
            
            # while os.path.exists(path):
            #     counter_exp +=1
            #     path = init_path + 'experiment_{:06d}'.format(counter_exp) + '/'
            # os.makedirs(path)
        
        print('The following experiment will have the given id : {}'.format(path[-2]))
        return path

    def reset_logs(self):
        self.ros_step = 0

        # Reseting robot time steps
        self.robot_time_list = list()
        self.ros_time_list = list()
        self.ros_step_list = list()

        ## Reseting positions
        self.robot_position_list = list()
        self.robot_yaw_position_list = list()
        self.robot_position_robot_frame_list = list()
        self.robot_yaw_robot_frame_list = list()

        # Reseting velocities
        self.robot_linear_velocity_list = list()
        self.robot_yaw_velocity_list = list()
        self.command_linear_velocity_list = list()
        self.command_yaw_velocity_list = list()

        # Reseting local map static
        self.local_map_static_list = list()
        self.local_map_static_rl_list = list()

        ## Reseting Global Map
        self.global_map_static = None
        self.global_map_static_rl = None
        self.global_map_size = 0

        # Reseting Social Cost 
        self.robot_social_cost_list = list()
        
        # Reseting Position of Humans
        self.humans_pose_list  = list()
        self.humans_visible_list = list()
        self.groups_id_list = list()
        self.groups_pose_list = list()
        self.humans_group_id_list = list()

        # Reseting Intermediate Goal Waypoint
        self.intermediate_waypoint_list  = list()

        # Reseting Final Goal
        self.final_goal_list = list()

        # Resetting Inference time
        self.time_inference_rl_list = list()
        self.time_inference_mpc_list = list()


    def store_measurements(self):
        print('Ongoing measurements')
        # Storing ros
        self.ros_step_list.append(self.ros_step)

        # Storing ros time
        self.ros_time_list.append(rospy.Time.now())

        # Storing robot time
        self.robot_time_list.append((self.timestamp_odometry_cb - self.time_start).to_sec())

        ## Storing Robot Position
        # Position from odometry
        self.robot_position_list.append([self.robot_x_position, self.robot_y_position])
        self.robot_yaw_position_list.append(self.robot_z_angular_position)

        # Position from TF
        self.robot_yaw_robot_frame_list.append(self.robot_yaw_robot_frame)
        self.robot_position_robot_frame_list.append(self.robot_position_robot_frame)

        # Velocity 
        self.robot_linear_velocity_list.append([self.robot_x_linear_velocity, self.robot_y_linear_velocity])
        self.robot_yaw_velocity_list.append(self.robot_z_angular_velocity)

        self.hri_data = self.ros_4_hri_interface.get_data()
        ## Storing Human Position 
        for human_id, human_data in self.hri_data[0].items():
            human_id = int(human_id[1:])
            humans_id_list = self.humans_id.tolist()
            if human_id in humans_id_list:
                index = humans_id_list.index(human_id)
            elif self.human_id_default in humans_id_list:
                index = humans_id_list.index(self.human_id_default)
            else:
                rospy.logerr("No more space in state.humans_ arrays. Number max is {}. It is fixed because of the use of SharedMemory multiprocessing".format(self.max_humans_world))
            self.humans_id[index] = human_id
            if len(human_data) >= 3:
                self.humans_group_id[index] = int(human_data[2][3:])  # 0: transform, 1:person_id, 2:group_id if group

            human_pose = human_data[0].transform.translation
            human_quat = human_data[0].transform.rotation
            x_pos, y_pos = human_pose.x, human_pose.y
            (_, _, yaw) = tf.transformations.euler_from_quaternion([
                human_quat.x,
                human_quat.y,
                human_quat.z,
                human_quat.w])

            self.humans_pose[index, :] = [x_pos, y_pos, self.constraint_angle(yaw)]

        self.humans_pose_list.append(self.humans_pose)

        ## Storing Group Position
        for group_id, group_data in self.hri_data[1].items():
            group_id  = int(group_id[3:])
            groups_id_list = self.groups_id.tolist()
            if group_id in groups_id_list:
                index = groups_id_list.index(group_id)
            elif self.group_id_default in groups_id_list:
                index = groups_id_list.index(self.group_id_default)
            else:
                rospy.logerr("No more space in state.groups_ arrays. Number max is {}. It is fixed because of the use of SharedMemory multiprocessing".format(self.max_groups_world))
            self.groups_id[index] = group_id
            self.groups_visible[index] = 0.  # how to get it ?

            group_pose = group_data[0].transform.translation
            x_pos, y_pos = group_pose.x, group_pose.y
            self.groups_pose[index, :] = [x_pos, y_pos]

        self.groups_pose_list.append(self.groups_pose)
        self.groups_id_list.append(self.groups_id)
        self.humans_visible_list.append(self.humans_visible)
        self.humans_group_id_list.append(self.humans_group_id)

        ## Storing Local Map
        self.local_map_static_list.append(self.local_map_static)
        self.local_map_static_rl_list.append(self.local_map_static_rl)

        ## Storing Global Map
        # Saved only once because it does not change

        ## Storing Speed Command
        self.command_linear_velocity_list.append(self.cmd_lin_vel)
        self.command_yaw_velocity_list.append(self.cmd_ang_vel)

        ## Storing Social Cost 
        self.robot_social_cost_list.append(self.robot_social_cost)

        ## Storing Intermediate Goal Waypoint
        self.intermediate_waypoint_list.append([self.x_wp, self.y_wp, self.yaw_wp])

        ## Storing Final Goal
        self.final_goal_list.append([self.x_final_goal, self.y_final_goal, self.yaw_final_goal])

        # Storing inference time
        self.time_inference_rl_list.append(self.time_inference_rl)
        self.time_inference_mpc_list.append(self.time_inference_mpc)

    def log_measurement(self):
        r''' Log the measurement into the given path '''

        path = self.path + 'repetition_{:06d}'.format(self.repetition_counter) + '/data'
        # Logging robot time
        np.save(path+'/robot_time', np.array(self.robot_time_list))
        
        # Logging ros time
        np.save(path+'/ros_time', np.array(self.ros_time_list))

        # Logging ros step
        np.save(path+'/ros_step', np.array(self.ros_step_list))

        ## Logging position - From odometry
        np.save(path+'/robot_position_odom', np.array(self.robot_position_list))
        np.save(path+'/robot_yaw_position_odom', np.array(self.robot_yaw_position_list))

        ## Logging position - From TF
        np.save(path+'/robot_position_robot_frame', np.array(self.robot_position_robot_frame_list))
        np.save(path+'/robot_yaw_robot_frame', np.array(self.robot_yaw_robot_frame_list))

        # Logging real velocity
        np.save(path+'/robot_velocity_odom', np.array(self.robot_linear_velocity_list))
        np.save(path+'/robot_yaw_velocity_odom', np.array(self.robot_yaw_velocity_list))
        
        # Logging command velocity
        np.save(path+'/command_linear_velocity', np.array(self.command_linear_velocity_list))
        np.save(path+'/command_angular_velocity', np.array(self.command_yaw_velocity_list))

        # Logging local map static
        np.save(path+'/local_map_static', self.local_map_static_list)
        np.save(path+'/local_map_static_rl', self.local_map_static_rl_list)

        ## Logging Global Map
        if self.global_map_static is not None :
            np.save(path+'/global_map_static', self.global_map_static)
            np.save(path+'/global_map_static_rl', self.global_map_static_rl)
            np.save(path+'/global_map_size', self.global_map_size)

        # Logging Social Cost 
        np.save(path+'/robot_social_cost', self.robot_social_cost_list)
        
        # Logging Position of Humans and Groups
        np.save(path+'/humans_position', np.array(self.humans_pose_list))
        np.save(path+'/groups_position', np.array(self.groups_pose_list))
        np.save(path+'/groups_id', np.array(self.groups_id_list))
        np.save(path+'/humans_visible', np.array(self.humans_visible_list))
        np.save(path+'/humans_group_id', np.array(self.humans_group_id_list))
        
        # Logging Intermediate Goal Waypoint
        np.save(path+'/intermediate_waypoint', np.array(self.intermediate_waypoint_list))

        # Logging Final Goal
        np.save(path+'/final_goal', np.array(self.final_goal_list))

        # Logging inference time
        np.save(path+'/time_inference_rl', np.array(self.time_inference_rl_list))
        np.save(path+'/time_inference_mpc', np.array(self.time_inference_mpc_list))

    def _callback_vel_cmd(self, data):
        self.cmd_lin_vel, self.cmd_ang_vel = data.linear.x, data.angular.z
    
    def _callback_time_inference_rl(self, data):
        self.time_inference_rl = data.data
    
    def _callback_time_inference_mpc(self, data):
        self.time_inference_mpc = data.data

    def get_robot_pose(self):
        r""" Function that compute the position of the robot on the map """
        
        try:
            map_2_robot = self.tf_listener.lookupTransform(self.map_frame, self.robot_frame, self.timestamp_tracking_latest)
        except:
            try:
                map_2_robot = self.tf_listener.lookupTransform(self.map_frame, self.robot_frame, rospy.Time(0))
                rospy.logwarn("Could not get transformation between {} and {} at correct time. Using latest available.".format(self.map_frame, self.robot_frame))
            except:
                rospy.logerr("Could not get transformation between {} and {}.".format(self.map_frame, self.robot_frame))
                return None
        x, y, _ = map_2_robot[0]
        (_, _, yaw) = tf.transformations.euler_from_quaternion([map_2_robot[1][0],
                                                                map_2_robot[1][1],
                                                                map_2_robot[1][2],
                                                                map_2_robot[1][3]])
        self.robot_yaw_robot_frame = float(self.constraint_angle(yaw))
        self.robot_position_robot_frame = np.array([x,y])
    

    def constraint_angle(self, angle, min_value=-np.pi, max_value=np.pi):
        """Constrains the given angle between -pi and pi.
        
        Note : function inspired from : https://gitlab.inria.fr/spring/wp6_robot_behavior/multiparty_interaction_simulator/-/blob/main/mpi_sim/utils/measurements.py """

        length = max_value - min_value
        new_angle = np.where(angle > max_value, min_value + ((angle - max_value) % length), angle)
        new_angle = np.where(angle < min_value, max_value - ((min_value - angle) % length), new_angle)

        return new_angle

    def _pub_vel_callback(self, event):
        r''' Function that checks the velocities in the memory and send it to the robot
         
          Warning : This functions locks the reading of the velocitites while sending it to the robot '''
        with self.lock_action_publishing :
            self.lin_vel, self.ang_vel = self._check_joint_vel_limits(self.lin_vel, self.ang_vel)
            twist_msg = Twist(Vector3(self.lin_vel, 0, 0), Vector3(0, 0, self.ang_vel))
            self._cmd_vel_pub.publish(twist_msg)
        # if self.lin_vel:
        #     rospy.loginfo("vel to robot : {}, {}".format(self.lin_vel, self.ang_vel))
        

    def handle_measurements(self, req):
        if req.start and not self.stop_measurement_flag :
            self.start_measurement_flag = True
            self.stop_measurement_flag = False
        if not req.start and self.start_measurement_flag :
            self.stop_measurement_flag = True
            self.start_measurement_flag = False
        return measurementResponse(req.start)

    def get_waypoint_pose(self):
        try:
            map_2_waypoint = self.tf_listener.lookupTransform(self.map_frame, self.intermediate_waypoint_goal_frame, rospy.Time(0))
            #rospy.logwarn("Could not get transformation between {} and {} at correct time. Using latest available.".format(self.map_frame, self.robot_frame))
        except:
            rospy.logerr("Could not get transformation between {} and {}.".format(self.map_frame, self.intermediate_waypoint_goal_frame))
            return False
        self.x_wp, self.y_wp, _ = map_2_waypoint[0]
        (_, _, yaw) = tf.transformations.euler_from_quaternion([map_2_waypoint[1][0],
                                                                map_2_waypoint[1][1],
                                                                map_2_waypoint[1][2],
                                                                map_2_waypoint[1][3]])
        self.yaw_wp = float(self.constraint_angle(yaw))
    
    def get_final_goal_pose(self):
        
        try:
            map_2_waypoint = self.tf_listener.lookupTransform(self.map_frame, self.final_goal_frame, rospy.Time(0))
            #rospy.logwarn("Could not get transformation between {} and {} at correct time. Using latest available.".format(self.map_frame, self.robot_frame))
        except:
            rospy.logerr("Could not get transformation between {} and {}.".format(self.map_frame, self.final_goal_frame))
            return False
        self.x_final_goal, self.y_final_goal, _ = map_2_waypoint[0]
        (_, _, yaw) = tf.transformations.euler_from_quaternion([map_2_waypoint[1][0],
                                                                map_2_waypoint[1][1],
                                                                map_2_waypoint[1][2],
                                                                map_2_waypoint[1][3]])
        self.yaw_final_goal = float(self.constraint_angle(yaw))
    
    def _odometry_callback(self, data):
        r''' Function that computes the velocity of the robot from the data '''
        self.odometry_data = data
    
        self.timestamp_odometry_cb = self.odometry_data.header.stamp

        # # Position data
        self.robot_x_position = self.odometry_data.pose.pose.position.x
        self.robot_y_position = self.odometry_data.pose.pose.position.y
        self.robot_z_position = self.odometry_data.pose.pose.position.z
        self.robot_x_angular_position = self.odometry_data.pose.pose.orientation.x
        self.robot_y_angular_position = self.odometry_data.pose.pose.orientation.y
        self.robot_z_angular_position = self.odometry_data.pose.pose.orientation.z
        self.robot_w_angular_position = self.odometry_data.pose.pose.orientation.w

        # Velocity data
        self.robot_x_linear_velocity = self.odometry_data.twist.twist.linear.x
        self.robot_y_linear_velocity = self.odometry_data.twist.twist.linear.y
        self.robot_z_linear_velocity = self.odometry_data.twist.twist.linear.z
        self.robot_x_angular_velocity = self.odometry_data.twist.twist.angular.x
        self.robot_y_angular_velocity = self.odometry_data.twist.twist.angular.y
        self.robot_z_angular_velocity = self.odometry_data.twist.twist.angular.z


    def _global_map_callback(self, data):
        r''' Function that computes the global map from the data and store it into the memory under self.global_map_static '''

        self.global_map_data = data
        self.x_global_map = self.global_map_data.info.origin.position.x
        self.y_global_map = self.global_map_data.info.origin.position.y
        self.global_map_width = self.global_map_data.info.width
        self.global_map_height = self.global_map_data.info.height
        self.global_map_resolution = self.global_map_data.info.resolution
        self.global_map_size = [[self.x_global_map, self.x_global_map + self.global_map_width*self.global_map_resolution],[self.y_global_map, self.y_global_map + self.global_map_height*self.global_map_resolution]]
        self.last_shape_global_map = (self.global_map_height, self.global_map_width)
        self.global_map_static = np.flip((np.asarray(self.global_map_data.data) / 100).reshape(self.last_shape_global_map), axis=0)
        

    def _global_map_rl_callback(self, data):
        r''' Function that computes the global map from the data and store it into the memory under self.global_map_static '''

        self.global_map_data = data
        self.x_global_map = self.global_map_data.info.origin.position.x
        self.y_global_map = self.global_map_data.info.origin.position.y
        self.global_map_width = self.global_map_data.info.width
        self.global_map_height = self.global_map_data.info.height
        self.global_map_resolution = self.global_map_data.info.resolution
        self.global_map_size = [[self.x_global_map, self.x_global_map + self.global_map_width*self.global_map_resolution],[self.y_global_map, self.y_global_map + self.global_map_height*self.global_map_resolution]]
        self.last_shape_global_map = (self.global_map_height, self.global_map_width)
        self.global_map_static_rl = np.flip((np.asarray(self.global_map_data.data) / 100).reshape(self.last_shape_global_map), axis=0)


    def _local_map_callback(self, data):
        r''' Function that computes the local map from the data and store it into the memory under self.local_map_static '''


        self.local_map_data = data
        self.timestamp_local_map_cb = self.local_map_data.header.stamp
        self.x_local_map = self.local_map_data.info.origin.position.x
        self.y_local_map = self.local_map_data.info.origin.position.y
        self.local_map_width = self.local_map_data.info.width
        self.local_map_height = self.local_map_data.info.height
        self.local_map_resolution = self.local_map_data.info.resolution
        self.local_map_size = [[self.x_local_map, self.x_local_map + self.local_map_width*self.local_map_resolution],[self.y_local_map, self.y_local_map + self.local_map_height*self.local_map_resolution]]
        self.last_shape_local_map = (self.local_map_height, self.local_map_width)
        self.local_map_static = (np.asarray(self.local_map_data.data) / 100).reshape(self.last_shape_local_map)

    def _local_map_rl_callback(self, data):
        r''' Function that computes the local map from the data and store it into the memory under self.local_map_static '''

        self.local_map_data = data
        self.timestamp_local_map_cb = self.local_map_data.header.stamp
        self.x_local_map = self.local_map_data.info.origin.position.x
        self.y_local_map = self.local_map_data.info.origin.position.y
        self.local_map_width = self.local_map_data.info.width
        self.local_map_height = self.local_map_data.info.height
        self.local_map_resolution = self.local_map_data.info.resolution
        self.local_map_size = [[self.x_local_map, self.x_local_map + self.local_map_width*self.local_map_resolution],[self.y_local_map, self.y_local_map + self.local_map_height*self.local_map_resolution]]
        self.last_shape_local_map = (self.local_map_height, self.local_map_width)
        self.local_map_static_rl = (np.asarray(self.local_map_data.data) / 100).reshape(self.last_shape_local_map)
    
    def _local_social_map_callback(self, data):
        r''' Function that computes the local map from the data and store it into the memory under self.local_static_map '''

        self.local_social_map_data = data
        self.timestamp_local_social_map_cb = self.local_social_map_data.header.stamp
        self.x_local_social_map = self.local_social_map_data.info.origin.position.x
        self.y_local_social_map = self.local_social_map_data.info.origin.position.y
        self.local_social_map_width = self.local_social_map_data.info.width
        self.local_social_map_height = self.local_social_map_data.info.height
        self.local_social_map_resolution = self.local_social_map_data.info.resolution
        self.local_social_map_size = [[self.x_local_social_map, self.x_local_social_map + self.local_social_map_width*self.local_social_map_resolution],[self.y_local_social_map, self.y_local_social_map + self.local_social_map_height*self.local_social_map_resolution]]
        self.last_shape_local_social_map = (self.local_social_map_height, self.local_social_map_width)
        self.local_social_map = (np.asarray(self.local_social_map_data.data) / 100).reshape(self.last_shape_local_social_map)
        # The social cost of the robot is the cost in the middle of the local cost map (where the robot is)
        self.robot_social_cost = self.local_social_map[np.shape(self.local_social_map)[0]//2][np.shape(self.local_social_map)[1]//2]
    

    def shutdown(self):
        rospy.loginfo("Stopping the BehaviorGenerator")
        self.close()
        rospy.loginfo("Killing the BehaviorGenerator node")

    def close(self):
        if self.measurement_server :
            self.measurement_server.shutdown()

        if self._subscribers:
            for subscriber in self._subscribers:
                subscriber.unregister()
        if self._publishers:
            for publisher in self._publishers:
                if isinstance(publisher, dict):
                    for pub in publisher.values():
                        pub.unregister()
                else:
                    publisher.unregister()
        if self._timers:
            for timer in self._timers:
                timer.shutdown()
        self.ros_4_hri_interface.close()


    # Method that checks

    def _check_all_sensors_ready(self):
        rospy.logdebug("START ALL SENSORS READY")
        self._check_rtabmap_ready()
        self._check_global_map_ready()
        self._check_local_map_ready()
        rospy.logdebug("ALL SENSORS READY")

    def _check_global_map_ready(self):
        self.global_map_data = None
        rospy.logdebug("Waiting for {} to be READY...".format(self.global_occupancy_map_topic))
        while self.global_map_data is None and not rospy.is_shutdown():
            try:
                self.global_map_data = rospy.wait_for_message(self.global_occupancy_map_topic, OccupancyGrid, timeout=5.0)
                rospy.logdebug("Current {} READY=>".format(self.global_occupancy_map_topic))

            except:
                rospy.logerr("Current {} not ready yet, retrying for getting global map".format(self.global_occupancy_map_topic))
        return self.global_map_data

    def _check_local_map_ready(self):
        self.local_map_data = None
        rospy.logdebug("Waiting for {} to be READY...".format(self.local_occupancy_map_topic))
        while self.local_map_data is None and not rospy.is_shutdown():
            try:
                self.local_map_data = rospy.wait_for_message(self.local_occupancy_map_topic, OccupancyGrid, timeout=5.0)
                rospy.logdebug("Current {} READY=>".format(self.local_occupancy_map_topic))

            except:
                rospy.logerr("Current {} not ready yet, retrying for getting local map".format(self.local_occupancy_map_topic))
        return self.local_map_data


    def _check_joint_states_ready(self):
        self.joint_states_data = None
        rospy.logdebug("Waiting for /joint_states to be READY...")
        while self.joint_states_data is None and not rospy.is_shutdown():
            try:
                self.joint_states_data = rospy.wait_for_message("/joint_states", JointState, timeout=5.0)
                rospy.logdebug("Current /joint_states READY=>")

            except:
                rospy.logerr("Current /joint_states not ready yet, retrying for getting joint states")
        if not self.get_indexes:
            # joint_states : head_1_joint, head_2_joint, wheel_left_joint, wheel_right_joint, caster_left_joint, caster_right_joint
            self.get_indexes = True

        return self.joint_states_data
    
    def _check_rtabmap_ready(self):
        rospy.logdebug("Waiting for rtabmap pose to be READY...")
        while self.rtabmap_ready is None and not rospy.is_shutdown():
            try:
                self.tf_listener.waitForTransform(self.map_frame, self.robot_frame, rospy.Time(0), rospy.Duration(5.0))
                self.rtabmap_ready = True
                rospy.logdebug("Current rtabmap pose READY=>")

            except:
                rospy.logerr("Current rtabmap pose not ready yet, retrying for getting rtabmap pose")
        return self.rtabmap_ready