#!/usr/bin/env python3
import numpy as np
import pkg_resources
import tf
import os
import yaml
import cv2
import sys
import rospy
from nav_msgs.msg import OccupancyGrid, MapMetaData
from robot_behavior.utils.behavior_global_path_planner import GlobalPathPlanner
from robot_behavior.utils.utils import ControllerConfig
import time
import actionlib
from robot_behavior.utils.utils import constraint_angle
from spring_msgs.msg import GlobalPlannerAction, GlobalPlannerFeedback, GlobalPlannerResult


class GlobalPathPlannerActionServer:
    def __init__(self, name):
        self._action_name = name

        mpc_config_name = rospy.get_param('~mpc_config_name', 'None')
        mpc_config_path = None if mpc_config_name == 'None' else mpc_config_name
        self.mpc_config_path = mpc_config_path
        self.namespace_slam = rospy.get_param('~namespace_slam', '/slam')
        self.map_frame = rospy.get_param('~map_frame', 'map')
        self.robot_frame = rospy.get_param('~robot_frame', 'base_footprint')
        self.global_occupancy_map_topic = rospy.get_param('~global_occupancy_map_topic', '/slam/global_map_rl_navigation')
        self.timer_error_max_time = rospy.get_param('~timer_error_max_time', 5)  # seconds
        self.final_goal_frame = rospy.get_param('~final_goal_frame', 'final_goal')
        self.intermediate_waypoint_goal_frame = rospy.get_param('~intermediate_waypoint_goal_frame', 'intermediate_waypoint_goal')
        self.global_path_loop_time = rospy.get_param('~global_path_loop_time', 2)

        self._feedback = GlobalPlannerFeedback()
        self._result = GlobalPlannerResult()
        
        self.read_config(filename=self.mpc_config_path)
        self.global_path_planner = GlobalPathPlanner(
            waypoint_resolution_in_meters=self.controller_config.orientation_distance_threshold,
            threshold_distance_orientation=self.controller_config.orientation_distance_threshold,
            threshold_distance_arrival=self.controller_config.waypoint_distance)

        self.global_map_data = None
        self.rtabmap_ready = False
        self.global_map_static = None

        self.final_goal = None

        self.goal_position = [5, 1]

        self.x_wp = 0.
        self.y_wp = 0.
        self.x_final = 0.
        self.y_final = 0.
        self.yaw_final = 0.
        self.yaw_wp = 0.
        self.flag_waypoint_okay = False
        self.tf_robot_pose_error = False
        self.tf_final_goal_pose_error = False

        self.robot_x_position = 0.
        self.robot_y_position = 0.
        self.robot_yaw = 0.
        self.robot_position = [0, 0]
        self.robot_position_former = [0, 0]
    
        self.init_ros_subscriber_and_publicher()

        self.tf_broadcaster = tf.TransformBroadcaster()  # Publish Transform to look for human
        self.tf_listener = tf.TransformListener()  # Listen Transform to look for human

        self._check_all_sensors_ready()

        rospy.on_shutdown(self.shutdown)

        self._a_server = actionlib.SimpleActionServer(self._action_name, GlobalPlannerAction, execute_cb=self.execute_cb, auto_start=False)
        self._a_server.start()

        rospy.loginfo("{} Initialization Ended".format(self._action_name))


    def execute_cb(self, goal):
        # helper variables
        rate = rospy.Rate(1/self.global_path_loop_time)
        self.init_time = rospy.Time.now()
        self.current_time = (rospy.Time.now() - self.init_time).to_sec()
        self.final_goal = None

        # publish info to the console for the user
        rospy.loginfo('Goal message : {}'.format(goal))
        self.final_goal = goal.final_goal

        # At time step 0, self.final_goal_frame = robot position
        self.get_robot_pose()
        # self.publish_tf_go_to_goals()
        self.x_final = self.robot_x_position
        self.y_final = self.robot_y_position
        self.yaw_final = self.robot_yaw
        self.quat_final = tf.transformations.quaternion_from_euler(0, 0, self.yaw_final)
        self.x_wp = 0.
        self.y_wp = 0.
        self.yaw_wp = 0.
        self.quat_wp = tf.transformations.quaternion_from_euler(0, 0, self.yaw_wp)
        self.flag_waypoint_okay = False
        self.tf_robot_pose_error = False
        self.tf_final_goal_pose_error = False
        self.timer_tf_final_goal_pose_error = None
        self.timer_tf_robot_pose_error = None

        # start executing the action
        while not rospy.is_shutdown():
            if self._a_server.is_preempt_requested():
                self._a_server.set_preempted()
                break

            # Reset TransformListener because cache_time not available in TransformListener
            # self.tf_listener = tf.TransformListener()

            self.current_time = (rospy.Time.now() - self.init_time).to_sec()
            # create messages that are used to publish feedback/result
            self._feedback = GlobalPlannerFeedback()
            self._result = GlobalPlannerResult()

            # publish the feedback
            self._publish_feedback(goal)
            rate.sleep()

            # if self.dist_to_goto_target < self.success_dist_thresh and self.ang_to_goto_target < self.success_ang_thresh:
            #     rospy.loginfo('%s: Succeeded' % self._action_name)
            #     self._a_server.set_succeeded(self._result)
            #     break

            # if self.controller_status_data.status == ControllerStatus.IS_FAILURE:
            #     text = "Controller failure"
            #     rospy.loginfo('{}: Aborted. {}'.format(self._action_name, text))
            #     self._a_server.set_aborted(self._result, text=text)
            #     break

            if self.current_time > goal.timeout:
                text = "Timeout"
                rospy.loginfo('{}: Aborted. {}'.format(self._action_name, text))
                self._a_server.set_aborted(self._result)
                break
            
            if self.tf_robot_pose_error:
                if not self.timer_tf_robot_pose_error:
                    self.timer_tf_robot_pose_error = rospy.Time.now()
                error_time = (rospy.Time.now() - self.timer_tf_robot_pose_error).to_sec()
                rospy.logwarn("Could not get transformation between {} and {}, no transformation available for {} seconds".format(self.map_frame, self.robot_frame, error_time))
                if error_time > self.timer_error_max_time:
                    text = "Could not get transformation between {} and {}, no transformation available during more than {} seconds. So the {} will be aborted".format(self.map_frame, self.robot_frame, self.timer_error_max_time, self._action_name)
                    rospy.loginfo('{}: Aborted. {}'.format(self._action_name, text))
                    self._a_server.set_aborted(self._result)
                    break

            if self.tf_final_goal_pose_error:
                if not self.timer_tf_final_goal_pose_error:
                    self.timer_tf_final_goal_pose_error = rospy.Time.now()
                error_time = (rospy.Time.now() - self.timer_tf_final_goal_pose_error).to_sec()
                rospy.logwarn("Could not get transformation between {} and {}, no transformation available for {} seconds".format(self.map_frame, self.final_goal_frame, error_time))
                if error_time > self.timer_error_max_time:
                    text = "Could not get transformation between {} and {}, no transformation available during more than {} seconds. So the {} will be aborted".format(self.map_frame, self.final_goal_frame, self.timer_error_max_time, self._action_name)
                    rospy.loginfo('{}: Aborted. {}'.format(self._action_name, text))
                    self._a_server.set_aborted(self._result)
                    break
            
            self.step(self.current_time)

        self._result_msg(goal)


    def _result_msg(self, goal):
        self.current_time = (rospy.Time.now() - self.init_time).to_sec()
        self._result.final_time = self.current_time


    def _publish_feedback(self, goal):
        self._feedback.current_time = self.current_time
        if self.flag_waypoint_okay:
            self._feedback.intermediate_waypoint.position.x = self.x_wp
            self._feedback.intermediate_waypoint.position.y = self.y_wp
            self._feedback.intermediate_waypoint.orientation.x = self.quat_wp[0]
            self._feedback.intermediate_waypoint.orientation.y = self.quat_wp[1]
            self._feedback.intermediate_waypoint.orientation.z = self.quat_wp[2]
            self._feedback.intermediate_waypoint.orientation.w = self.quat_wp[3]
        self._a_server.publish_feedback(self._feedback)
    

    def step(self, init_time):
        r''' Step the ros wrapper'''

        # Localizing the robot in the environment
        if not self.get_robot_pose():
            self.tf_robot_pose_error = True
        else:
            self.tf_robot_pose_error = False
            self.timer_tf_robot_pose_error = None

        # Get the self.final_goal_frame tf
        if not self.get_final_goal_pose():
            self.tf_final_goal_pose_error = True
        else:
            self.tf_final_goal_pose_error = False
            self.timer_tf_final_goal_pose_error = None
        
        rospy.logdebug('tf_robot_pose_error: {}'.format(self.tf_robot_pose_error))
        rospy.logdebug('timer_tf_robot_pose_error: {}'.format(self.timer_tf_robot_pose_error))
        rospy.logdebug('tf_final_goal_pose_error: {}'.format(self.tf_final_goal_pose_error))
        rospy.logdebug('timer_tf_final_goal_pose_error: {}'.format(self.timer_tf_final_goal_pose_error))

        if self.global_map_static is not None:
            waypoint_target = self.global_path_planner.run(
                                                            global_map = self.global_map_static, #cv2.dilate(self.global_map_static,kernel=cv2.getStructuringElement(cv2.MORPH_RECT,(8,8)),iterations = 1),
                                                            target_pose = [self.x_final, self.y_final, self.yaw_final],
                                                            robot_pose=[self.robot_x_position, self.robot_y_position, self.robot_yaw]
                            )
            if waypoint_target:
                self.flag_waypoint_okay = True
                self.x_wp, self.y_wp, self.yaw_wp = waypoint_target
                self.quat_wp = tf.transformations.quaternion_from_euler(0, 0, self.yaw_wp)
                rospy.logdebug('Robot position : x {:.2f} \t y {:.2f} \t yaw {:.2f}'.format(self.robot_x_position, self.robot_y_position, self.robot_yaw))
                rospy.logdebug('Goal position : x {:.2f} \t y {:.2f} \t yaw {:.2f}'.format(self.x_final, self.y_final, self.yaw_final))
                rospy.logdebug('Intermediate waypoint position : x {:.2f} \t y {:.2f} \t yaw {:.2f}'.format(self.x_wp, self.y_wp, self.yaw_wp))
            else:
                self.flag_waypoint_okay = False
                rospy.logwarn("The waypoint was not computed properly")

            if self.global_path_planner.global_map_with_waypoints is not None :
                self.publish_global_waypoint_map(self.global_path_planner.global_map_with_waypoints)
            else :
                rospy.logwarn('waypoint map not ready !') 

            self.robot_position_former  = self.robot_position
            
            self.publish_tf_go_to_goals()


    def read_config(self, filename=None):
        if filename is None:
            filename = pkg_resources.resource_filename(
                'social_mpc', 'config/social_mpc.yaml')
        elif os.path.isfile(filename):
            self.passed_config_loaded = True
        else:
            filename = pkg_resources.resource_filename(
                'social_mpc', 'config/social_mpc.yaml')
        config = yaml.load(open(filename), Loader=yaml.FullLoader)
        self.controller_config = ControllerConfig(config)

        self.goal_finder_enabled = self.controller_config.goal_finder_enabled
        self.path_planner_enabled = self.controller_config.path_planner_enabled
        self.update_goals_enabled = self.controller_config.update_goals_enabled


    def init_ros_subscriber_and_publicher(self):
        r''' Initialize the subscribers and publishers '''
        self._subscribers = []
        self._publishers = []
        self._action_server = []
        self._timers = []

        ### ROS Subscribers
        self._subscribers.append(rospy.Subscriber(self.global_occupancy_map_topic, OccupancyGrid, callback=self._global_map_callback, queue_size=1))

        ### ROS Publishers
        # self._cmd_vel_pub = rospy.Publisher('/cmd_vel', Twist, queue_size=1)
        # self._publishers.append(self._cmd_vel_pub)

        ### Publishing map in Rviz
        self._global_waypoint_map_pub = rospy.Publisher(self.namespace_slam + '/global_waypoint_map', OccupancyGrid, queue_size=10, latch=True)
        self._publishers.append(self._global_waypoint_map_pub)

        rospy.loginfo("Initializing the %s " % self._action_name)

        # self._check_publishers_connection()


    def publish_tf_go_to_goals(self):
        current_time = rospy.Time.now()
        # to debug 
        # quat_final = tf.transformations.quaternion_from_euler(0, 0, self.yaw_final)
        # self.x_final, self.y_final = self.goal_position
        # self.tf_broadcaster.sendTransform(
        #     (self.x_final, self.y_final, 0),
        #     quat_final,
        #     current_time,
        #     self.final_goal_frame,
        #     self.map_frame
        # )
        if self.flag_waypoint_okay:
            self.tf_broadcaster.sendTransform(
                (self.x_wp, self.y_wp, 0),
                self.quat_wp,
                current_time,
                self.intermediate_waypoint_goal_frame,
                self.map_frame
            )


    def get_final_goal_pose(self):
        try:
            map_2_final_goal = self.tf_listener.lookupTransform(self.map_frame, self.final_goal_frame, rospy.Time(0))
            #rospy.logwarn("Could not get transformation between {} and {} at correct time. Using latest available.".format(self.map_frame, self.robot_frame))
        except:
            #rospy.logerr("Could not get transformation between {} and {}.".format(self.map_frame, self.robot_frame))
            return False
        self.x_final, self.y_final, _ = map_2_final_goal[0]
        (_, _, yaw) = tf.transformations.euler_from_quaternion([map_2_final_goal[1][0],
                                                                map_2_final_goal[1][1],
                                                                map_2_final_goal[1][2],
                                                                map_2_final_goal[1][3]])
        self.yaw_final = float(constraint_angle(yaw))
        return True


    def get_robot_pose(self):
        r""" Function that compute the position of the robot on the map """
        # try:
        #     map_2_robot = self.tf_listener.lookupTransform(self.map_frame, self.robot_frame, self.timestamp_tracking_latest)
        # except:
        try:
            map_2_robot = self.tf_listener.lookupTransform(self.map_frame, self.robot_frame, rospy.Time(0))
            #rospy.logwarn("Could not get transformation between {} and {} at correct time. Using latest available.".format(self.map_frame, self.robot_frame))
        except:
            #rospy.logerr("Could not get transformation between {} and {}.".format(self.map_frame, self.robot_frame))
            return False
        self.robot_x_position, self.robot_y_position, _ = map_2_robot[0]
        (_, _, yaw) = tf.transformations.euler_from_quaternion([map_2_robot[1][0],
                                                                map_2_robot[1][1],
                                                                map_2_robot[1][2],
                                                                map_2_robot[1][3]])
        self.robot_yaw = float(constraint_angle(yaw))
        self.robot_position = np.array([self.robot_x_position, self.robot_y_position])
        return True


    def publish_global_waypoint_map(self, map):
        self._global_waypoint_map_pub.publish(self.set_global_map_object(map))


    def set_global_map_object(self, map):
        occupancy_grid = OccupancyGrid()
        occupancy_grid.info = MapMetaData()
        occupancy_grid.info.map_load_time = rospy.Time.now()
        occupancy_grid.info.resolution =  self.global_map_resolution
        occupancy_grid.info.width = self.global_map_width
        occupancy_grid.info.height = self.global_map_height
        occupancy_grid.info.origin = self.origin_global_map
        
        occupancy_grid.data = (np.flip(map, axis=0)*100).flatten().astype(np.int8)

        occupancy_grid.header.stamp = self.timestamp_global_map_cb
        occupancy_grid.header.frame_id = self.map_frame

        return occupancy_grid


    def _global_map_callback(self, data):
        r''' Function that computes the global map from the data and store it into the memory under self.global_map_static '''
        if data:
            self.global_map_data = data
            self.timestamp_global_map_cb = self.global_map_data.header.stamp
            self.x_global_map = self.global_map_data.info.origin.position.x
            self.y_global_map = self.global_map_data.info.origin.position.y
            self.origin_global_map = self.global_map_data.info.origin
            self.global_map_width = self.global_map_data.info.width
            self.global_map_height = self.global_map_data.info.height
            self.global_map_resolution = self.global_map_data.info.resolution
            self.global_map_size = [[self.x_global_map, self.x_global_map + self.global_map_width*self.global_map_resolution],[self.y_global_map, self.y_global_map + self.global_map_height*self.global_map_resolution]]
            self.last_shape_global_map = (self.global_map_height, self.global_map_width)
            self.global_map_static = np.flip((np.asarray(self.global_map_data.data) / 100).reshape(self.last_shape_global_map), axis=0)
            if self.global_path_planner.state_config is None :
                self.global_path_planner.set_state_config(global_map_size=self.global_map_size,
                                        global_map_height=self.global_map_height,
                                        global_map_width=self.global_map_width,
                                        global_map_scale=1/0.05
                                        )


    def _check_all_sensors_ready(self):
        rospy.logdebug("START ALL SENSORS READY")
        self._check_rtabmap_ready()
        self._check_global_map_ready()
        rospy.logdebug("ALL SENSORS READY")


    def _check_global_map_ready(self):
        self.global_map_data = None
        rospy.logdebug("Waiting for {} to be READY...".format(self.global_occupancy_map_topic))
        while self.global_map_data is None and not rospy.is_shutdown():
            try:
                self.global_map_data = rospy.wait_for_message(self.global_occupancy_map_topic, OccupancyGrid, timeout=5.0)
                rospy.logdebug("Current {} READY=>".format(self.global_occupancy_map_topic))

            except:
                rospy.logerr("Current {} not ready yet, retrying for getting global map".format(self.global_occupancy_map_topic))
        return self.global_map_data


    def _check_rtabmap_ready(self):
        rospy.logdebug("Waiting for rtabmap pose to be READY...")
        while self.rtabmap_ready is None and not rospy.is_shutdown():
            try:
                self.tf_listener.waitForTransform(self.map_frame, self.robot_frame, rospy.Time(0), rospy.Duration(5.0))
                self.rtabmap_ready = True
                rospy.logdebug("Current rtabmap pose READY=>")

            except:
                rospy.logerr("Current rtabmap pose not ready yet, retrying for getting rtabmap pose")
        return self.rtabmap_ready


    def shutdown(self):
        rospy.loginfo("Stopping the %s " % self._action_name)
        self.close()
        rospy.loginfo("Killing the %s node" % self._action_name)


    def close(self):
        if self._subscribers:
            for subscriber in self._subscribers:
                subscriber.unregister()

        if self._publishers:
            for publisher in self._publishers:
                if isinstance(publisher, dict):
                    for pub in publisher.values():
                        pub.unregister()
                else:
                    publisher.unregister()
        if self._timers:
            for timer in self._timers:
                timer.shutdown()

        if self._action_server:
            self._action_server.shutdown()
