#!/usr/bin/env python3
import actionlib
import pkg_resources
import tf
import sys
import os
import yaml
import numpy as np
import rospy
from nav_msgs.msg import OccupancyGrid
from robot_behavior.utils.behavior_goal_finder import GoalFinder
from robot_behavior.utils.utils import constraint_angle, RobotState, ControllerConfig, RobotConfig
import time
from robot_behavior.utils.ros_4_hri_interface import ROS4HRIInterface 
from spring_msgs.msg import GoalFinderAction, GoalFinderFeedback, GoalFinderResult


class GoalFinderActionServer:
    def __init__(self, name):
        self._action_name = name

        mpc_config_name = rospy.get_param('~mpc_config_name', 'None')
        # robot_config_name = rospy.get_param('~robot_config_name', 'None')
        mpc_config_path = None if mpc_config_name == 'None' else mpc_config_name
        # robot_config_path = None if robot_config_name == 'None' else robot_config_name
        self.mpc_config_path = mpc_config_path
        # self.robot_config_path = robot_config_path
        # self.namespace_slam = rospy.get_param('~namespace_slam', '/slam')
        self.map_frame = rospy.get_param('~map_frame', 'map')
        self.robot_frame = rospy.get_param('~robot_frame', 'base_footprint')
        self.global_occupancy_map_topic = rospy.get_param('~global_occupancy_map_topic', '/slam/global_map_rl_navigation')
        self.max_humans_world = rospy.get_param('max_humans_world', 20)
        self.max_groups_world = rospy.get_param('max_groups_world', 10)
        self.namespace = rospy.get_param('namespace', 'behavior_generator')
        self.final_goal_frame = rospy.get_param('~final_goal_frame', 'final_goal')
        self.timer_error_max_time = rospy.get_param('~timer_error_max_time', 5)  # seconds
        self.goal_finder_loop_time = rospy.get_param('~goal_finder_loop_time', 2)

        self._feedback = GoalFinderFeedback()
        self._result = GoalFinderResult()
        
        # self.read_robot_config(filename=self.robot_config_path)
        self.read_config(filename=self.mpc_config_path)
        self.goal_finder = GoalFinder(
            controller_config=self.controller_config)

        self.global_map_data = None
        self.rtabmap_ready = False
        self.global_map_static = None
        self.hri_data = None

        self.human_id_default = -1
        self.group_id_default = -1

        self.group_id = None
        self.body_id = None
        self.person_id = None
        self.wrong_group_id = False
        self.wrong_body_id = False
        self.wrong_person_id = False

        self.x_final = 0.
        self.y_final = 0.
        self.yaw_final = 0.
        self.quat_final = None
        self.flag_final_goal_okay = False
        self.tf_robot_pose_error = False
        self.timer_tf_robot_pose_error = None
        self.timer_wrong_id_error = None

        self.robot_x_position = 0.
        self.robot_y_position = 0.
        self.robot_yaw = 0.
        self.robot_position = [0, 0]
    
        self.init_ros_subscriber_and_publicher()

        self.tf_broadcaster = tf.TransformBroadcaster()  # Publish Transform to look for human
        self.tf_listener = tf.TransformListener()  # Listen Transform to look for human

        self._check_all_sensors_ready()

        self.ros_4_hri_interface = ROS4HRIInterface()
        self.init_state()

        rospy.on_shutdown(self.shutdown)

        self._a_server = actionlib.SimpleActionServer(self._action_name, GoalFinderAction, execute_cb=self.execute_cb, auto_start=False)
        self._a_server.start()

        rospy.loginfo("{} Initialization Ended".format(self._action_name))


    def execute_cb(self, goal):
        # helper variables
        rate = rospy.Rate(1/self.goal_finder_loop_time)
        self.init_time = rospy.Time.now()
        self.current_time = (rospy.Time.now() - self.init_time).to_sec()
        self.group_id = None
        self.body_id = None
        self.person_id = None

        # publish info to the console for the user
        rospy.loginfo('Goal message : {}'.format(goal))
        self.group_id = goal.group_id
        self.body_id = goal.body_id
        self.person_id = goal.person_id

        self.x_final = 0.
        self.y_final = 0.
        self.yaw_final = 0.
        self.quat_final = None
        self.flag_final_goal_okay = False
        self.wrong_group_id = False
        self.wrong_body_id = False
        self.wrong_person_id = False
        self.tf_robot_pose_error = False
        self.timer_tf_robot_pose_error = None
        self.timer_wrong_id_error = None

        # start executing the action
        while not rospy.is_shutdown():
            if self._a_server.is_preempt_requested():
                self._a_server.set_preempted()
                break

            # Reset TransformListener because cache_time not available in TransformListener
            # self.tf_listener = tf.TransformListener()

            self.current_time = (rospy.Time.now() - self.init_time).to_sec()
            # create messages that are used to publish feedback/result
            self._feedback = GoalFinderFeedback()
            self._result = GoalFinderResult()

            # publish the feedback
            self._publish_feedback(goal)
            rate.sleep()

            if self.current_time > goal.timeout:
                text = "Timeout"
                rospy.loginfo('{}: Aborted. {}'.format(self._action_name, text))
                self._a_server.set_aborted(self._result)
                break
                        
            if self.wrong_body_id:
                if not self.timer_wrong_id_error:
                    self.timer_wrong_id_error = rospy.Time.now()
                error_time = (rospy.Time.now() - self.timer_wrong_id_error).to_sec()
                rospy.logwarn('body_id {} not in the list of the detected bodies for {} seconds'.format(self.body_id, error_time))
                if error_time > self.timer_error_max_time:
                    text = "body_id {} not in the list of the detected bodies during more than {} seconds. So the {} will be aborted".format(self.body_id, self.timer_error_max_time, self._action_name)
                    rospy.loginfo('{}: Aborted. {}'.format(self._action_name, text))
                    self._a_server.set_aborted(self._result)
                    break
            
            if self.wrong_group_id:
                if not self.timer_wrong_id_error:
                    self.timer_wrong_id_error = rospy.Time.now()
                error_time = (rospy.Time.now() - self.timer_wrong_id_error).to_sec()
                rospy.logwarn('group_id {} not in the list of the detected groups for {} seconds'.format(self.group_id, error_time))
                if error_time > self.timer_error_max_time:
                    text = "group_id {} not in the list of the detected groups during more than {} seconds. So the {} will be aborted".format(self.group_id, self.timer_error_max_time, self._action_name)
                    rospy.loginfo('{}: Aborted. {}'.format(self._action_name, text))
                    self._a_server.set_aborted(self._result)
                    break
            
            if self.wrong_person_id:
                if not self.timer_wrong_id_error:
                    self.timer_wrong_id_error = rospy.Time.now()
                error_time = (rospy.Time.now() - self.timer_wrong_id_error).to_sec()
                rospy.logwarn('person_id {} not in the list of the detected persons for {} seconds'.format(self.person_id, error_time))
                if error_time > self.timer_error_max_time:
                    text = "person_id {} not in the list of the detected persons during more than {} seconds. So the {} will be aborted".format(self.person_id, self.timer_error_max_time, self._action_name)
                    rospy.loginfo('{}: Aborted. {}'.format(self._action_name, text))
                    self._a_server.set_aborted(self._result)
                    break

            if self.tf_robot_pose_error:
                if not self.timer_tf_robot_pose_error:
                    self.timer_tf_robot_pose_error = rospy.Time.now()
                error_time = (rospy.Time.now() - self.timer_tf_robot_pose_error).to_sec()
                rospy.logwarn("Could not get transformation between {} and {}, no transformation available for {} seconds".format(self.map_frame, self.robot_frame, error_time))
                if error_time > self.timer_error_max_time:
                    text = "Could not get transformation between {} and {}, no transformation available during more than {} seconds. So the {} will be aborted".format(self.map_frame, self.robot_frame, self.timer_error_max_time, self._action_name)
                    rospy.loginfo('{}: Aborted. {}'.format(self._action_name, text))
                    self._a_server.set_aborted(self._result)
                    break
            
            self.step(self.current_time)

        self._result_msg(goal)


    def _result_msg(self, goal):
        self.current_time = (rospy.Time.now() - self.init_time).to_sec()
        self._result.final_time = self.current_time


    def _publish_feedback(self, goal):
        self._feedback.current_time = self.current_time
        if self.flag_final_goal_okay:
            self._feedback.final_goal.position.x = self.x_final
            self._feedback.final_goal.position.y = self.y_final
            self._feedback.final_goal.orientation.x = self.quat_final[0]
            self._feedback.final_goal.orientation.y = self.quat_final[1]
            self._feedback.final_goal.orientation.z = self.quat_final[2]
            self._feedback.final_goal.orientation.w = self.quat_final[3]
        self._a_server.publish_feedback(self._feedback)
    

    def step(self, init_time):
        r''' Step the ros wrapper'''

        # Localizing the robot in the environment
        # Localizing the robot in the environment
        if not self.get_robot_pose():
            self.tf_robot_pose_error = True
        else:
            self.tf_robot_pose_error = False
            self.timer_tf_robot_pose_error = None

        self.hri_data = self.ros_4_hri_interface.get_data()
        self.recreate_state()

        if not self.group_id and not self.body_id and not self.person_id:
            self.wrong_group_id = True
            self.wrong_body_id = True
            self.wrong_person_id = True

        if self.group_id:
            go_to_group_target = int(self.group_id[3:])
            if float(go_to_group_target) in self.state.groups_id.tolist() and go_to_group_target >= 0:
                self.shared_goto_target_group_id = [1, go_to_group_target]
                self.wrong_group_id = False
                self.timer_wrong_id_error = None
            else:
                self.wrong_group_id = True
        
        if self.body_id:
            go_to_body_target = int(self.body_id[1:])
            if float(go_to_body_target) in self.state.humans_id.tolist() and go_to_body_target >= 0:
                self.shared_goto_target_human_id = [1, go_to_body_target]
                self.wrong_body_id = False
                self.timer_wrong_id_error = None
            else:
                self.wrong_body_id = True
        
        if self.person_id:
            go_to_person_target = self.person_id
            if go_to_person_target in self.hri_data[2].keys():
                go_to_person_target = int(self.hri_data[2][self.person_id][1:])
                if float(go_to_person_target) in self.state.humans_id.tolist():
                    self.shared_goto_target_human_id = [1, go_to_person_target]
                    self.wrong_person_id = False
                    self.timer_wrong_id_error = None
                else:
                    self.wrong_person_id = True
            else:
                self.wrong_person_id = True

        if self.shared_goto_target_human_id[0] or self.shared_goto_target_group_id[0]:
            # final_goal = None
            final_goal = self.goal_finder.run(
                self.shared_humans,
                self.shared_groups,
                self.state.robot_pose,
                self.shared_goto_target_human_id,
                self.shared_goto_target_group_id,
                self.global_map_static,
            )
            if final_goal:
                self.flag_final_goal_okay = True
                self.x_final, self.y_final, self.yaw_final = final_goal
                self.quat_final = tf.transformations.quaternion_from_euler(0, 0, self.yaw_final)
                rospy.logdebug('Robot position : x {:.2f} \t y {:.2f} \t yaw {:.2f}'.format(self.robot_x_position, self.robot_y_position, self.robot_yaw))
                rospy.logdebug('Goal position : x {:.2f} \t y {:.2f} \t yaw {:.2f}'.format(self.x_final, self.y_final, self.yaw_final))
            else:
                self.flag_final_goal_okay = False
                rospy.logwarn("The {} was not computed properly".format(self.final_goal_frame))

        self.publish_tf_go_to_goals()


    def read_config(self, filename=None):
        if filename is None:
            filename = pkg_resources.resource_filename(
                'social_mpc', 'config/social_mpc.yaml')
        elif os.path.isfile(filename):
            self.passed_config_loaded = True
        else:
            filename = pkg_resources.resource_filename(
                'social_mpc', 'config/social_mpc.yaml')
        config = yaml.load(open(filename), Loader=yaml.FullLoader)
        self.controller_config = ControllerConfig(config)

        self.goal_finder_enabled = self.controller_config.goal_finder_enabled
        self.path_planner_enabled = self.controller_config.path_planner_enabled
        self.update_goals_enabled = self.controller_config.update_goals_enabled
    

    def read_robot_config(self, filename=None):
        if filename is None:
            filename = pkg_resources.resource_filename(
                'sim2d', 'config/robot.yaml')
            rospy.logdebug("No filename provided for the robot configuration, basic robot config loaded")
        elif os.path.isfile(filename):
            rospy.logdebug("Desired robot config loaded")
        else:
            filename = pkg_resources.resource_filename(
                'sim2d', 'config/robot.yaml')
            rospy.logdebug("Desired filename for the robot configuration does not exist, basic robot config loaded")
        config = yaml.load(open(filename), Loader=yaml.FullLoader)
        self.robot_config = RobotConfig(config)


    def init_ros_subscriber_and_publicher(self):
        r''' Initialize the subscribers and publishers '''
        self._subscribers = []
        self._publishers = []
        self._action_server = []
        self._timers = []

        ### ROS Subscribers
        self._subscribers.append(rospy.Subscriber(self.global_occupancy_map_topic, OccupancyGrid, callback=self._global_map_callback, queue_size=1))

        ### ROS Publishers
        # self._cmd_vel_pub = rospy.Publisher('/cmd_vel', Twist, queue_size=1)
        # self._publishers.append(self._cmd_vel_pub)

        rospy.loginfo("Initializing the %s " % self._action_name)

        # self._check_publishers_connection()


    def publish_tf_go_to_goals(self):
        if self.flag_final_goal_okay:
            current_time = rospy.Time.now()
            self.tf_broadcaster.sendTransform(
                (self.x_final, self.y_final, 0),
                self.quat_final,
                current_time,
                self.final_goal_frame,
                self.map_frame
            )


    def get_robot_pose(self):
        r""" Function that compute the position of the robot on the map """
        try:
            map_2_robot = self.tf_listener.lookupTransform(self.map_frame, self.robot_frame, rospy.Time(0))
            #rospy.logwarn("Could not get transformation between {} and {} at correct time. Using latest available.".format(self.map_frame, self.robot_frame))
        except:
            #rospy.logerr("Could not get transformation between {} and {}.".format(self.map_frame, self.robot_frame))
            return False
        self.robot_x_position, self.robot_y_position, _ = map_2_robot[0]
        (_, _, yaw) = tf.transformations.euler_from_quaternion([map_2_robot[1][0],
                                                                map_2_robot[1][1],
                                                                map_2_robot[1][2],
                                                                map_2_robot[1][3]])
        self.robot_yaw = float(constraint_angle(yaw))
        self.robot_position = np.array([self.robot_x_position, self.robot_y_position])
        return True


    def _global_map_callback(self, data):
        r''' Function that computes the global map from the data and store it into the memory under self.global_map_static '''
        if data:
            self.global_map_data = data
            self.x_global_map = self.global_map_data.info.origin.position.x
            self.y_global_map = self.global_map_data.info.origin.position.y
            self.global_map_width = self.global_map_data.info.width
            self.global_map_height = self.global_map_data.info.height
            self.global_map_resolution = self.global_map_data.info.resolution
            self.global_map_size = [[self.x_global_map, self.x_global_map + self.global_map_width*self.global_map_resolution],[self.y_global_map, self.y_global_map + self.global_map_height*self.global_map_resolution]]
            self.last_shape_global_map = (self.global_map_height, self.global_map_width)
            self.global_map_static = np.flip((np.asarray(self.global_map_data.data) / 100).reshape(self.last_shape_global_map), axis=0)
            if self.goal_finder.state_config is None :
                self.goal_finder.set_state_config(global_map_size=self.global_map_size,
                                                    global_map_height=self.global_map_height,
                                                    global_map_width=self.global_map_width,
                                                    global_map_scale=1/0.05
                                                    )


    def reset_state(self):
        nb_humans = self.max_humans_world
        self.state.humans_id = np.ones(nb_humans) * self.human_id_default
        self.state.humans_group_id = np.ones(nb_humans) * self.group_id_default
        self.state.groups_id = np.ones(self.max_groups_world) * self.group_id_default
        self.state.humans_visible = np.ones(nb_humans)
        self.state.groups_visible = np.ones(self.max_groups_world)
        self.state.humans_pose = np.zeros((nb_humans, 3))  # x, y, angle
        self.state.groups_pose = np.zeros((self.max_groups_world, 2))  # x, y
        self.state.humans_velocity = np.zeros((nb_humans, 2))
        self.state.robot_pose = np.zeros(3)
        self.state.robot_velocity = np.zeros(2)

        self.shared_humans = np.zeros((nb_humans, 7))  # x, y, orientation, lin vel, ang vel, group id, id
        self.shared_groups = np.zeros((self.max_groups_world, 3))  # x, y, id
        self.shared_goto_target_group_id = np.zeros(2)  # flag, id
        self.shared_goto_target_human_id = np.zeros(2)  # flag, id


    def recreate_state(self):
        self.reset_state()

        self.state.robot_pose = np.array([self.robot_x_position, self.robot_y_position, self.robot_yaw])
        
        for human_id, human_data in self.hri_data[0].items():
            human_id = int(human_id[1:])
            humans_id_list = self.state.humans_id.tolist()
            if human_id in humans_id_list:
                index = humans_id_list.index(human_id)
            elif self.human_id_default in humans_id_list:
                index = humans_id_list.index(self.human_id_default)
            else:
                rospy.logerr("No more space in state.humans_ arrays. Number max is {}. It is fixed because of the use of SharedMemory multiprocessing".format(self.max_humans_world))
            self.state.humans_id[index] = human_id
            self.state.humans_visible[index] = 0.  # how to get it ?
            if len(human_data) >= 3:
                self.state.humans_group_id[index] = int(human_data[2][3:])  # 0: transform, 1:person_id, 2:group_id if group

            human_pose = human_data[0].transform.translation
            human_quat = human_data[0].transform.rotation
            x_pos, y_pos = human_pose.x, human_pose.y
            (_, _, yaw) = tf.transformations.euler_from_quaternion([
                human_quat.x,
                human_quat.y,
                human_quat.z,
                human_quat.w])

            lin = 0.  # how to get it ?
            ang = 0.  # how to get it ?

            self.state.humans_pose[index, :] = [x_pos, y_pos, constraint_angle(yaw)]
            self.state.humans_velocity[index, :] = [lin, ang]

        rospy.logdebug('humans_id : {}'.format(self.state.humans_id.tolist()))

        for group_id, group_data in self.hri_data[1].items():
            group_id  = int(group_id[3:])
            groups_id_list = self.state.groups_id.tolist()
            if group_id in groups_id_list:
                index = groups_id_list.index(group_id)
            elif self.group_id_default in groups_id_list:
                index = groups_id_list.index(self.group_id_default)
            else:
                rospy.logerr("No more space in state.groups_ arrays. Number max is {}. It is fixed because of the use of SharedMemory multiprocessing".format(self.max_groups_world))
            self.state.groups_id[index] = group_id
            self.state.groups_visible[index] = 0.  # how to get it ?

            group_pose = group_data[0].transform.translation
            x_pos, y_pos = group_pose.x, group_pose.y
            self.state.groups_pose[index, :] = [x_pos, y_pos]

        rospy.logdebug('groups_id ; {}'.format(self.state.groups_id.tolist()))

        self.shared_humans[:, -1] = -1
        idx = np.where(self.state.humans_visible == 0)[0]
        n = len(idx)
        self.shared_humans[:n, -1] = self.state.humans_id[idx]
        self.shared_humans[:n, -2] = self.state.humans_group_id[idx]
        self.shared_humans[:n, :3] = self.state.humans_pose[idx, :]
        self.shared_humans[:n, 3:5] = self.state.humans_velocity[idx, :]

        self.shared_groups[:, -1] = -1
        idx = np.where(self.state.groups_visible == 0)[0]
        n = len(idx)
        self.shared_groups[:n, :2] = self.state.groups_pose[idx, :]
        self.shared_groups[:n, -1] = self.state.groups_id[idx]
    

    def init_state(self):
        self.state = RobotState()
        self.list_attr = ["robot_pose",
                          "robot_velocity",
                          "groups_pose",
                          "humans_pose",
                          "humans_velocity",
                          "humans_visible",
                          "groups_visible",
                          "groups_id",
                          "humans_group_id",
                          "humans_id",
                          ]
        for attr in self.list_attr:
            self.state.__setattr__(attr, None)

        self.shared_humans = None
        self.shared_groups = None
        self.shared_goto_target_group_id = None
        self.shared_goto_target_human_id = None

        self.reset_state()


    def _check_all_sensors_ready(self):
        rospy.logdebug("START ALL SENSORS READY")
        self._check_rtabmap_ready()
        self._check_global_map_ready()
        rospy.logdebug("ALL SENSORS READY")


    def _check_global_map_ready(self):
        self.global_map_data = None
        rospy.logdebug("Waiting for {} to be READY...".format(self.global_occupancy_map_topic))
        while self.global_map_data is None and not rospy.is_shutdown():
            try:
                self.global_map_data = rospy.wait_for_message(self.global_occupancy_map_topic, OccupancyGrid, timeout=5.0)
                rospy.logdebug("Current {} READY=>".format(self.global_occupancy_map_topic))

            except:
                rospy.logerr("Current {} not ready yet, retrying for getting global map".format(self.global_occupancy_map_topic))
        return self.global_map_data


    def _check_rtabmap_ready(self):
        rospy.logdebug("Waiting for rtabmap pose to be READY...")
        while self.rtabmap_ready is None and not rospy.is_shutdown():
            try:
                self.tf_listener.waitForTransform(self.map_frame, self.robot_frame, rospy.Time(0), rospy.Duration(5.0))
                self.rtabmap_ready = True
                rospy.logdebug("Current rtabmap pose READY=>")

            except:
                rospy.logerr("Current rtabmap pose not ready yet, retrying for getting rtabmap pose")
        return self.rtabmap_ready


    def shutdown(self):
        rospy.loginfo("Stopping the %s " % self._action_name)
        self.close()
        rospy.loginfo("Killing the %s node" % self._action_name)


    def close(self):
        if self._subscribers:
            for subscriber in self._subscribers:
                subscriber.unregister()

        if self._publishers:
            for publisher in self._publishers:
                if isinstance(publisher, dict):
                    for pub in publisher.values():
                        pub.unregister()
                else:
                    publisher.unregister()
        if self._timers:
            for timer in self._timers:
                timer.shutdown()

        if self._action_server:
            self._action_server.shutdown()
        self.ros_4_hri_interface.close()
