#!/usr/bin/env python3
import rospy
import actionlib
from spring_msgs.msg import (GoToPosition,
                             GoToEntity,
                             LookAtEntity,
                             LookAtPosition,
                             GoalFinderAction,
                             GoalFinderGoal,
                             GlobalPlannerAction,
                             GlobalPlannerGoal,
                             LocalPlannerAction,
                             LocalPlannerGoal,
                             ControllerStatus,
                             LookAtEntityAction,
                             LookAtPositionAction,
                             GoToEntityAction,
                             GoToPositionAction)
from actionlib_msgs.msg import GoalStatus
from trajectory_msgs.msg import JointTrajectory, JointTrajectoryPoint
from sensor_msgs.msg import JointState
from robot_behavior.utils.utils import constraint_angle, local_to_global, ControllerConfig, RobotConfig
import tf
import os
import sys
import yaml
import time
import numpy as np
from robot_behavior.utils.ros_4_hri_interface import ROS4HRIInterface
import pkg_resources


class PlanningManagerNode:
    def __init__(self):
        robot_config_name = rospy.get_param('~robot_config_name', 'None')
        robot_config_path = None if robot_config_name == 'None' else robot_config_name
        self.robot_config_path = robot_config_path
        mpc_config_name = rospy.get_param('~mpc_config_name', 'None')
        mpc_config_path = None if mpc_config_name == 'None' else mpc_config_name
        self.mpc_config_path = mpc_config_path
        self._goal_finder_action_name = rospy.get_param('~goal_finder_action', 'goal_finder_action')
        self._global_path_planner_action_name = rospy.get_param('~global_path_planner_action', 'global_path_planner_action')
        self._mpc_local_path_planner_action_name = rospy.get_param('~mpc_local_path_planner_action', 'mpc_local_path_planner_action')
        self._rl_local_path_planner_action_name = rospy.get_param('~rl_local_path_planner_action', 'rl_local_path_planner_action')
        self._look_at_body_action_name = rospy.get_param('~look_at_body_action', 'look_at_body_action')
        self._look_at_person_action_name = rospy.get_param('~look_at_person_action', 'look_at_person_action')
        self._look_at_group_action_name = rospy.get_param('~look_at_group_action', 'look_at_group_action')
        self._look_at_position_action_name = rospy.get_param('~look_at_position_action', 'look_at_position_action')
        self._go_to_body_action_name = rospy.get_param('~go_to_body_action', 'go_to_body_action')
        self._go_to_person_action_name = rospy.get_param('~go_to_person_action', 'go_to_person_action')
        self._go_to_group_action_name = rospy.get_param('~go_to_group_action', 'go_to_group_action')
        self._go_to_position_action_name = rospy.get_param('~go_to_position_action', 'go_to_position_action')
        self.namespace = rospy.get_param('namespace', 'behavior_generator')
        self.map_frame = rospy.get_param('~map_frame', 'map')
        self.robot_frame = rospy.get_param('~robot_frame', 'base_footprint')
        self.default_timeout = rospy.get_param('~default_timeout', 150)
        self.max_humans_world = rospy.get_param('max_humans_world', 20)
        self.max_groups_world = rospy.get_param('max_groups_world', 10)
        self.timer_error_max_time = rospy.get_param('~timer_error_max_time', 5)  # seconds
        self.final_goal_frame = rospy.get_param('~final_goal_frame', 'final_goal')
        self.pan_goal_frame = rospy.get_param('~pan_goal_frame', 'pan_goal')
        self.human_goal_frame = rospy.get_param('~human_goal_frame', 'human_goal')
        self.planning_manager_loop_time = rospy.get_param('~planning_manager_loop_time', 2)
        self._local_path_planner_name = rospy.get_param('~local_path_planner_name', 'social_mpc')

        self.read_robot_config(filename=self.robot_config_path)
        self.read_config(filename=self.mpc_config_path)
        
        self.ros_4_hri_interface = ROS4HRIInterface()
        self.reset_state()
        self.hri_data = None

        self.x_final = 0.
        self.y_final = 0.
        self.yaw_final = 0.
        self.robot_x_position = 0.
        self.robot_y_position = 0.
        self.robot_yaw = 0.
        self.robot_position = [0, 0]
        self.x_human_features = 0.
        self.y_human_features = 0.
        self.yaw_human_features = 0.
        self.flag_publish_final_goal = False
        self.flag_publish_pan_goal = False
        self.flag_publish_human_goal = False

        # continuous flags
        self.is_continuous_look_at = False
        self.continuous_look_at_done = False
        self.is_continuous_go_to = False
        self.continuous_go_to_done = False
        self.controller_status_msg = None
        self.timer_look_at_id_error = None
        self.timer_go_to_id_error = None
        self.look_at_body = False
        self.look_at_group = False
        self.look_at_person = False
        self.look_at_position = False
        self.timer_go_to_id_error = None
        self.go_to_body = False
        self.go_to_group = False
        self.go_to_person = False
        self.go_to_position = False

        self.go_to_running = False
        self.look_at_running = False

        self.goto_target_id = -1
        self.human_target_id = -1
        self.pan_target_id = -1

        # To get indexes of the joints
        self.joint_states_data = None
        self.get_indexes = False
        self.pan_ind = None
        self.tilt_ind = None
        self.current_pan = None
        self.current_tilt = None
        self.head_goal = False
        self.base_goal = False
        self.x_pan_target = 0.
        self.y_pan_target = 0.
        self.pan_angle_cmd = 0.  # to be set and publish
        self.tilt_angle_cmd = 0.  # to be set and publish
        self.lin_vel = 0.  # get value from action server feedback
        self.ang_vel = 0.  # get value from action server feedback

        self.setting_goals = False
        self._local_path_planner_action_client = None
        self._local_path_planner_action_name = None

        self.reset_go_to_data()
        self.reset_look_at_data()

        self.init_ros_subscriber_and_publicher()

        rospy.on_shutdown(self.shutdown)

        self._check_all_sensors_ready()

        rospy.loginfo("PlanningManagerNode Initialization Ended")

    
    def init_ros_subscriber_and_publicher(self):
        r''' Initialize the subscribers and publishers '''
        self._subscribers = []
        self._publishers = []
        self._action_server = []
        self._timers = []
        self._action_client = []

        ### ROS Subscribers
        self._subscribers.append(rospy.Subscriber('/joint_states', JointState, callback=self._joint_states_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.namespace + '/action/go_to_position', GoToPosition, callback=self._go_to_position_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.namespace + '/action/go_to_body', GoToEntity, callback=self._go_to_body_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.namespace + '/action/go_to_person', GoToEntity, callback=self._go_to_person_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.namespace + '/action/go_to_group', GoToEntity, callback=self._go_to_group_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.namespace + '/action/look_at_position', LookAtPosition, callback=self._look_at_position_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.namespace + '/action/look_at_body', LookAtEntity, callback=self._look_at_body_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.namespace + '/action/look_at_person', LookAtEntity, callback=self._look_at_person_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.namespace + '/action/look_at_group', LookAtEntity, callback=self._look_at_group_callback, queue_size=1))

        ### ROS Publishers
        # self._cmd_vel_pub = rospy.Publisher('/cmd_vel', Twist, queue_size=1)
        # self._publishers.append(self._cmd_vel_pub)
        self._controller_status_pub = rospy.Publisher(self.namespace + '/controller_status', ControllerStatus, queue_size=1)
        self._publishers.append(self._controller_status_pub)
        self._pan_vel_pub = rospy.Publisher('/head_controller/command', JointTrajectory, queue_size=1)
        self._publishers.append(self._pan_vel_pub)

        self._timers.append(rospy.Timer(rospy.Duration(self.controller_config.h), callback=self._pub_vel_callback))

        # Creates the SimpleActionClient
        self._goal_finder_action_client = actionlib.SimpleActionClient(self._goal_finder_action_name, GoalFinderAction)
        self._global_path_planner_action_client = actionlib.SimpleActionClient(self._global_path_planner_action_name, GlobalPlannerAction)
        self._mpc_local_path_planner_action_client = actionlib.SimpleActionClient(self._mpc_local_path_planner_action_name, LocalPlannerAction)
        self._rl_local_path_planner_action_client = actionlib.SimpleActionClient(self._rl_local_path_planner_action_name, LocalPlannerAction)
        self._look_at_body_action_client = actionlib.SimpleActionClient(self.namespace + '/' + self._look_at_body_action_name, LookAtEntityAction)
        self._look_at_group_action_client = actionlib.SimpleActionClient(self.namespace + '/' + self._look_at_group_action_name, LookAtEntityAction)
        self._look_at_person_action_client = actionlib.SimpleActionClient(self.namespace + '/' + self._look_at_person_action_name, LookAtEntityAction)
        self._look_at_position_action_client = actionlib.SimpleActionClient(self.namespace + '/' + self._look_at_position_action_name, LookAtPositionAction)
        self._go_to_body_action_client = actionlib.SimpleActionClient(self.namespace + '/' + self._go_to_body_action_name, GoToEntityAction)
        self._go_to_group_action_client = actionlib.SimpleActionClient(self.namespace + '/' + self._go_to_group_action_name, GoToEntityAction)
        self._go_to_person_action_client = actionlib.SimpleActionClient(self.namespace + '/' + self._go_to_person_action_name, GoToEntityAction)
        self._go_to_position_action_client = actionlib.SimpleActionClient(self.namespace + '/' + self._go_to_position_action_name, GoToPositionAction)
        self._action_client.append(self._goal_finder_action_client)
        self._action_client.append(self._global_path_planner_action_client)
        self._action_client.append(self._mpc_local_path_planner_action_client)
        # self._action_client.append(self._rl_local_path_planner_action_client)
        self._action_client.append(self._look_at_body_action_client)
        self._action_client.append(self._look_at_group_action_client)
        self._action_client.append(self._look_at_person_action_client)
        self._action_client.append(self._look_at_position_action_client)
        self._action_client.append(self._go_to_body_action_client)
        self._action_client.append(self._go_to_group_action_client)
        self._action_client.append(self._go_to_person_action_client)
        self._action_client.append(self._go_to_position_action_client)

        self.tf_broadcaster = tf.TransformBroadcaster()
        self.tf_listener = tf.TransformListener()

        # Waits until the action server has started up and started listening for goals
        rospy.logdebug('Waiting until all the action servers are ready and cancel all previous goals')
        for action_client in self._action_client:
            action_client.wait_for_server()
            action_client.cancel_all_goals()
        rospy.sleep(0.1)
        rospy.logdebug('All the action servers are READY')

        self.set_local_path_planner(local_path_planner_name=self._local_path_planner_name)

        self._check_publishers_connection()

        rospy.loginfo("Initializing the PlanningManagerNode")
    

    def set_local_path_planner(self, local_path_planner_name):
        if local_path_planner_name == 'social_mpc':
            self._local_path_planner_action_client = self._mpc_local_path_planner_action_client
            self._local_path_planner_action_name = self._mpc_local_path_planner_action_name
        elif local_path_planner_name == 'social_rl':
            self._local_path_planner_action_client = self._rl_local_path_planner_action_client
            self._local_path_planner_action_name = self._rl_local_path_planner_action_name
        else:
            rospy.logerr('Local Path Planner {} not implemented'.format(local_path_planner_name))

        self._local_path_planner_action_client.wait_for_server()
        # self._local_path_planner_action_client.cancel_all_goals()


    def run(self):
        r''' Runs the ros wrapper '''
        self.init_time = rospy.Time.now()
        flag = False
        rate = rospy.Rate(1/self.planning_manager_loop_time)

        while not rospy.is_shutdown():
            # Upfate HRI data
            self.hri_data = self.ros_4_hri_interface.get_data()
            self.reset_state()
            self.update_state()

            # Get pan and tilt angle
            self.current_pan = self.joint_states_data.position[self.pan_ind]
            self.current_tilt = self.joint_states_data.position[self.tilt_ind]
           
            # Localizing the robot in the environment
            self.get_robot_pose()

            # Get or publish the final_goal_frame tf
            if self.flag_publish_final_goal:
                self.publish_tf_go_to_goals()
            else:
                self.get_final_goal_pose()
            
            if self.flag_publish_pan_goal:
                self.publish_tf_look_at_goals()

            if self.flag_publish_human_goal:
                self.publish_tf_human_goal()
            
            self.publish_controller_status()           

            goal_finder_action_server_state = self._goal_finder_action_client.get_state()
            global_path_planner_action_server_state = self._global_path_planner_action_client.get_state()
            local_path_planner_action_server_state = self._local_path_planner_action_client.get_state()
            rospy.loginfo('global_path_planner_action_server_state: {}'.format(global_path_planner_action_server_state))
            rospy.loginfo('goal_finder_action_server_state: {}'.format(goal_finder_action_server_state))
            rospy.loginfo('local_path_planner_action_server_state: {}'.format(local_path_planner_action_server_state))

            if not self.setting_goals:
                if local_path_planner_action_server_state == GoalStatus.ABORTED:
                    # rospy.loginfo('Action server {} ABORTED'.format(self._local_path_planner_action_name))
                    # if local_planner aborted and global_path_planner is running, cancel action global_path_planner
                    if global_path_planner_action_server_state != GoalStatus.LOST:
                        self._global_path_planner_action_client.cancel_all_goals()
                    if not self.go_to_position:
                        if goal_finder_action_server_state != GoalStatus.LOST:
                            self._goal_finder_action_client.cancel_all_goals()

                if goal_finder_action_server_state == GoalStatus.ABORTED:
                    # rospy.loginfo('Action server {} ABORTED'.format(self._goal_finder_action_name))
                    # if goal_finder aborted and global_path_planner is running, cancel action global_path_planner
                    if not self.go_to_position:
                        if global_path_planner_action_server_state != GoalStatus.LOST:
                            self._global_path_planner_action_client.cancel_all_goals()
                        if local_path_planner_action_server_state != GoalStatus.LOST:
                            self._local_path_planner_action_client.cancel_all_goals()

                if global_path_planner_action_server_state == GoalStatus.ABORTED:
                    # rospy.loginfo('Action server {} ABORTED'.format(self._global_path_planner_action_name))
                    if local_path_planner_action_server_state != GoalStatus.LOST:
                        self._local_path_planner_action_client.cancel_all_goals()

                if goal_finder_action_server_state == GoalStatus.SUCCEEDED:
                    # rospy.loginfo('Action server {} SUCCEEDED'.format(self._goal_finder_action_name))
                    # if goal_finder succeded and global_path_planner is running, cancel action global_path_planner
                    if global_path_planner_action_server_state != GoalStatus.LOST:
                        self._global_path_planner_action_client.cancel_all_goals()
                    if local_path_planner_action_server_state != GoalStatus.LOST:
                        self._local_path_planner_action_client.cancel_all_goals()

                if global_path_planner_action_server_state == GoalStatus.SUCCEEDED:
                    # rospy.loginfo('Action server {} SUCCEEDED'.format(self._global_path_planner_action_name))
                    if local_path_planner_action_server_state != GoalStatus.LOST:
                        self._local_path_planner_action_client.cancel_all_goals()
                
                # if local_path_planner_action_server_state == GoalStatus.PREEMPTED:
                #     rospy.loginfo('Action server {} PREEMPTED'.format(self._local_path_planner_action_name))
                #     # if goal_finder preempted and global_path_planner is running, cancel action global_path_planner
                #     if not self.go_to_position:
                #         if goal_finder_action_server_state != GoalStatus.LOST and goal_finder_action_server_state != GoalStatus.PREEMPTED:
                #             self._goal_finder_action_client.cancel_all_goals()
                #     if global_path_planner_action_server_state != GoalStatus.LOST and global_path_planner_action_server_state != GoalStatus.PREEMPTED:
                #         self._global_path_planner_action_client.cancel_all_goals()

                if goal_finder_action_server_state == GoalStatus.PREEMPTED:
                    # rospy.loginfo('Action server {} PREEMPTED'.format(self._goal_finder_action_name))
                    # if goal_finder preempted and global_path_planner is running, cancel action global_path_planner
                    if not self.go_to_position:
                        if global_path_planner_action_server_state != GoalStatus.LOST and global_path_planner_action_server_state != GoalStatus.PREEMPTED:
                            self._global_path_planner_action_client.cancel_all_goals()
                        if local_path_planner_action_server_state != GoalStatus.LOST and local_path_planner_action_server_state != GoalStatus.PREEMPTED:
                            self._local_path_planner_action_client.cancel_all_goals()

                if global_path_planner_action_server_state == GoalStatus.PREEMPTED:
                    # rospy.loginfo('Action server {} PREEMPTED'.format(self._global_path_planner_action_name))
                    if local_path_planner_action_server_state != GoalStatus.LOST and local_path_planner_action_server_state != GoalStatus.PREEMPTED:
                        self._local_path_planner_action_client.cancel_all_goals()
                
                if self.pan_target_id != -1 and not self.look_at_position:
                    if self.look_at_body or self.look_at_person:
                        id_list = self.humans_id.tolist()
                    if self.look_at_group:
                        id_list = self.groups_id.tolist()
                    if float(self.pan_target_id) not in id_list:
                        if not self.timer_look_at_id_error:
                            self.timer_look_at_id_error = rospy.Time.now()
                        error_time = (rospy.Time.now() - self.timer_look_at_id_error).to_sec()
                        rospy.logwarn('Pan target id {} lost for {} seconds'.format(self.pan_target_id, error_time))
                        if error_time > self.timer_error_max_time:
                            # cancel all look_at goals by default because we do not know which one is running
                            rospy.logwarn('Pan target id {} lost for more than {} seconds. So the look_at action will be canceled'.format(self.pan_target_id, self.timer_error_max_time))
                            self._look_at_body_action_client.cancel_all_goals()
                            self._look_at_group_action_client.cancel_all_goals()
                            self._look_at_person_action_client.cancel_all_goals()
                            self.reset_look_at_data()
                    else:
                        self.timer_look_at_id_error = None

            rate.sleep()


    def _pub_vel_callback(self, event):
        if self.flag_publish_pan_goal:
            self.pan_angle_cmd = constraint_angle(np.arctan2(self.y_pan_target - self.robot_y_position, self.x_pan_target - self.robot_x_position) - self.robot_yaw) # to check
            self.publish_head_cmd(self.pan_angle_cmd, self.tilt_angle_cmd)


    def publish_head_cmd(self, pan, tilt):
        pan_angle, tilt_angle = self._check_joint_pos_limits(pan, tilt)
        
        head_msg = JointTrajectory()
        p = JointTrajectoryPoint()
        p.positions = [None] * 2
        p.positions[0] = pan_angle
        p.positions[1] = tilt_angle
        # head_msg.header.frame_id = "head"
        # head_msg.header.stamp = rospy.Time.now()
        head_msg.joint_names = [None]*2
        head_msg.joint_names[0] = "head_1_joint"  # pan
        head_msg.joint_names[1] = "head_2_joint"  # tilt
        head_msg.points = [None]
        head_msg.points[0] = p
        head_msg.points[0].time_from_start = rospy.Duration.from_sec(0.5)
        self._pan_vel_pub.publish(head_msg)
        rospy.loginfo('pan_angle goal : {}  tilt angle goal : {}'.format(pan_angle, tilt_angle))


    def reset_state(self):
        self.human_id_default = -1
        self.group_id_default = -1

        self.humans_id = np.ones(self.max_humans_world) * self.human_id_default
        self.groups_id = np.ones(self.max_groups_world) * self.group_id_default
        self.humans_pose = np.zeros((self.max_humans_world, 3))  # x, y, angle
        self.groups_pose = np.zeros((self.max_groups_world, 2))  # x, y
    

    def update_state(self):
        for human_id, human_data in self.hri_data[0].items():
            human_id = int(human_id[1:])
            humans_id_list = self.humans_id.tolist()
            if human_id in humans_id_list:
                index = humans_id_list.index(human_id)
            elif self.human_id_default in humans_id_list:
                index = humans_id_list.index(self.human_id_default)
            else:
                rospy.logerr("No more space in state.humans_ arrays. Number max is {}. It is fixed because of the use of SharedMemory multiprocessing".format(self.max_humans_world))
            self.humans_id[index] = human_id

            human_pose = human_data[0].transform.translation
            human_quat = human_data[0].transform.rotation
            x_pos, y_pos = human_pose.x, human_pose.y
            (_, _, yaw) = tf.transformations.euler_from_quaternion([
                human_quat.x,
                human_quat.y,
                human_quat.z,
                human_quat.w])

            self.humans_pose[index, :] = [x_pos, y_pos, constraint_angle(yaw)]

        rospy.logdebug('humans_id : {}'.format(self.humans_id.tolist()))

        for group_id, group_data in self.hri_data[1].items():
            group_id  = int(group_id[3:])
            groups_id_list = self.groups_id.tolist()
            if group_id in groups_id_list:
                index = groups_id_list.index(group_id)
            elif self.group_id_default in groups_id_list:
                index = groups_id_list.index(self.group_id_default)
            else:
                rospy.logerr("No more space in state.groups_ arrays. Number max is {}. It is fixed because of the use of SharedMemory multiprocessing".format(self.max_groups_world))
            self.groups_id[index] = group_id

            group_pose = group_data[0].transform.translation
            x_pos, y_pos = group_pose.x, group_pose.y
            self.groups_pose[index, :] = [x_pos, y_pos]
        
        rospy.logdebug('groups_id ; {}'.format(self.groups_id.tolist()))


    def publish_tf_go_to_goals(self):
        current_time = rospy.Time.now()
        quat_final = tf.transformations.quaternion_from_euler(0, 0, self.yaw_final)
        self.tf_broadcaster.sendTransform(
            (self.x_final, self.y_final, 0),
            quat_final,
            current_time,
            self.final_goal_frame,
            self.map_frame
        )


    def publish_tf_look_at_goals(self):
        current_time = rospy.Time.now()
        quat_pan_goal = tf.transformations.quaternion_from_euler(0, 0, self.robot_yaw)
        self.tf_broadcaster.sendTransform(
            (self.x_pan_target, self.y_pan_target, 0),
            quat_pan_goal,
            current_time,
            self.pan_goal_frame,
            self.map_frame
        )
    

    def publish_tf_human_goal(self):
        current_time = rospy.Time.now()
        quat = tf.transformations.quaternion_from_euler(0, 0, self.yaw_human_features)
        self.tf_broadcaster.sendTransform(
            (self.x_human_features, self.y_human_features, 0),
            quat,
            current_time,
            self.human_goal_frame,
            self.map_frame
        )


    def publish_controller_status(self):
        controller_status_msg = ControllerStatus()
        controller_status_msg.goto_target_id = self.goto_target_id
        controller_status_msg.human_target_id = self.human_target_id
        controller_status_msg.pan_target_id = self.pan_target_id

        controller_status_msg.is_continuous_look_at = self.is_continuous_look_at
        controller_status_msg.continuous_look_at_done = self.continuous_look_at_done
        controller_status_msg.is_continuous_go_to = self.is_continuous_go_to        
        controller_status_msg.continuous_go_to_done = self.continuous_go_to_done

        if controller_status_msg.goto_target_id != -1:
            dist_to_goto_target = np.linalg.norm(np.array([self.robot_x_position - self.x_final, self.robot_y_position - self.y_final]))
            controller_status_msg.dist_to_goto_target = [np.abs(self.robot_x_position - self.x_final),
                                              np.abs(self.robot_y_position - self.y_final),
                                              np.abs(constraint_angle(self.robot_yaw - self.yaw_final)),
                                              dist_to_goto_target]
        
        if controller_status_msg.human_target_id != -1:
            dist_to_human_target = np.linalg.norm(np.array([self.x_human_features, self.y_human_features]))
            controller_status_msg.dist_to_human_target = [np.abs(self.x_human_features),
                                               np.abs(self.y_human_features),
                                               np.abs(constraint_angle(self.yaw_human_features)),
                                               dist_to_human_target]

        if controller_status_msg.pan_target_id != -1:
            head_joint_errors = []
            if self.pan_ind is not None:
                head_joint_errors.append(np.abs(constraint_angle(np.arctan2(self.y_pan_target - self.robot_y_position, self.x_pan_target - self.robot_x_position) - constraint_angle(self.robot_yaw + self.current_pan))))
            if self.tilt_ind is not None:
                head_joint_errors.append(0.)
            controller_status_msg.head_joint_errors = head_joint_errors

        self.controller_status_msg = controller_status_msg
        self._controller_status_pub.publish(self.controller_status_msg)


    def get_final_goal_pose(self):
        try:
            map_2_final_goal = self.tf_listener.lookupTransform(self.map_frame, self.final_goal_frame, rospy.Time(0))
            #rospy.logwarn("Could not get transformation between {} and {} at correct time. Using latest available.".format(self.map_frame, self.robot_frame))
        except:
            #rospy.logerr("Could not get transformation between {} and {}.".format(self.map_frame, self.robot_frame))
            return False
        self.x_final, self.y_final, _ = map_2_final_goal[0]
        (_, _, yaw) = tf.transformations.euler_from_quaternion([map_2_final_goal[1][0],
                                                                map_2_final_goal[1][1],
                                                                map_2_final_goal[1][2],
                                                                map_2_final_goal[1][3]])
        self.yaw_final = float(constraint_angle(yaw))
        return True


    def get_robot_pose(self):
        r""" Function that compute the position of the robot on the map """
        # try:
        #     map_2_robot = self.tf_listener.lookupTransform(self.map_frame, self.robot_frame, self.timestamp_tracking_latest)
        # except:
        try:
            map_2_robot = self.tf_listener.lookupTransform(self.map_frame, self.robot_frame, rospy.Time(0))
            #rospy.logwarn("Could not get transformation between {} and {} at correct time. Using latest available.".format(self.map_frame, self.robot_frame))
        except:
            #rospy.logerr("Could not get transformation between {} and {}.".format(self.map_frame, self.robot_frame))
            return False
        self.robot_x_position, self.robot_y_position, _ = map_2_robot[0]
        (_, _, yaw) = tf.transformations.euler_from_quaternion([map_2_robot[1][0],
                                                                map_2_robot[1][1],
                                                                map_2_robot[1][2],
                                                                map_2_robot[1][3]])
        self.robot_yaw = float(constraint_angle(yaw))
        self.robot_position = np.array([self.robot_x_position, self.robot_y_position])
        return True


    def goal_finder_action_server(self, goal):
        # cancel all the current goal that are runnning on the action server
        self._goal_finder_action_client.cancel_all_goals()
        # the only way to not put the action server preempted in mode is to wait between cancel goal and send goal
        rospy.sleep(1)

        # Sends the goal to the action server
        self._goal_finder_action_client.send_goal(goal, feedback_cb=self.feedback_cb)

        # Waits for the server to finish performing the action
        # self._goal_finder_action_client.wait_for_result()

        # Prints out the result of executing the action
        # result = self._goal_finder_action_client.get_result()
        # return result


    def global_path_planner_action_server(self, goal):
        # cancel all the current goal that are runnning on the action server
        self._global_path_planner_action_client.cancel_all_goals()
        # the only way to not put the action server preempted in mode is to wait between cancel goal and send goal
        rospy.sleep(1)

        # Sends the goal to the action server
        self._global_path_planner_action_client.send_goal(goal, feedback_cb=self.feedback_cb)

        # Waits for the server to finish performing the action
        # self._global_path_planner_action_client.wait_for_result()

        # Prints out the result of executing the action
        # result = self._global_path_planner_action_client.get_result()
        # return result


    def local_path_planner_action_server(self, goal):
        # cancel all the current goal that are runnning on the action server
        self._local_path_planner_action_client.cancel_all_goals()
        # the only way to not put the action server preempted in mode is to wait between cancel goal and send goal
        rospy.sleep(1)

        local_path_planner_param = rospy.get_param('~local_path_planner_name', 'social_mpc')
        if local_path_planner_param != self._local_path_planner_name:
            self.set_local_path_planner(local_path_planner_name=local_path_planner_param)
            rospy.loginfo('{} action server activated'.format(self._local_path_planner_action_name))

        # Sends the goal to the action server
        self._local_path_planner_action_client.send_goal(goal, feedback_cb=self.feedback_cb)

        # Waits for the server to finish performing the action
        # self._local_path_planner_action_client.wait_for_result()

        # Prints out the result of executing the action
        # result = self._local_path_planner_action_client.get_result()
        # return result


    def feedback_cb(self, msg):
        # rospy.loginfo('Feedback received: {}'.format(msg))
        pass
    

    def reset_continuous_look_at_flag(self):
        self.is_continuous_look_at = False
        self.continuous_look_at_done = False


    def reset_continuous_go_to_flag(self):
        self.is_continuous_go_to = False
        self.continuous_go_to_done = False


    def reset_go_to_data(self):
        self.go_to_position_data = None
        self.go_to_body_data = None
        self.go_to_person_data = None
        self.go_to_group_data = None
        self.timer_go_to_id_error = None
        self.go_to_body = False
        self.go_to_group = False
        self.go_to_person = False
        self.go_to_position = False
        self.flag_publish_final_goal = False
        self.goto_target_id = -1
        self.x_final = 0.
        self.y_final = 0.
        self.yaw_final = 0.

    
    def reset_look_at_data(self):
        self.look_at_person_data = None
        self.look_at_group_data = None
        self.look_at_position_data = None
        self.look_at_body_data = None
        self.wrong_target = False
        self.look_at_done = False
        self.timer_look_at_id_error = None
        self.look_at_body = False
        self.look_at_group = False
        self.look_at_person = False
        self.look_at_position = False
        self.flag_publish_pan_goal = False
        self.pan_target_id = -1
        self.x_pan_target = 0.
        self.y_pan_target = 0.

    
    def _go_to_position_callback(self, data):
        self.reset_go_to_data()
        self.go_to_position_data = data
        go_to_position_target = np.asarray(data.go_to_target_pose)
        self.setting_goals = True
        if go_to_position_target[-1] == -1:
            # stop publishing final_goal_frame tf, stop global_path_planner and local_path_planner action servers
            self._global_path_planner_action_client.cancel_all_goals()
            self._local_path_planner_action_client.cancel_all_goals()
        elif go_to_position_target[-1] == 1:
                self.goto_target_id = 1  # is str[] so can not have id as int, 1 by default
                self.flag_publish_final_goal = True
                self.x_final = go_to_position_target[0]
                self.y_final = go_to_position_target[1]
                self.yaw_final = go_to_position_target[2]
                self.go_to_position = True                
                if go_to_position_target[-2] == 1:
                    # local frame
                    self.x_final, self.y_final = local_to_global(
                        robot_position=np.array([self.robot_x_position, self.robot_y_position, self.robot_yaw]), 
                        x=np.array([self.x_final, self.y_final])
                    )
                    self.go_to_position = True                    
                    goal = GlobalPlannerGoal()
                    # goal.final_goal =
                    goal.timeout = self.default_timeout
                    self.global_path_planner_action_server(goal=goal)
                    while self._global_path_planner_action_client.get_state() != GoalStatus.ACTIVE:
                        rospy.sleep(0.02)

                    goal = LocalPlannerGoal()
                    # goal.intermediate_waypoint =
                    goal.timeout = self.default_timeout
                    self.local_path_planner_action_server(goal=goal)
                    while self._local_path_planner_action_client.get_state() != GoalStatus.ACTIVE:
                        rospy.sleep(0.02)
                elif go_to_position_target[-2] == 0:
                    # global frame
                    self.go_to_position = True
                    goal = GlobalPlannerGoal()
                    # goal.final_goal =
                    goal.timeout = self.default_timeout
                    self.global_path_planner_action_server(goal=goal)
                    while self._global_path_planner_action_client.get_state() != GoalStatus.ACTIVE:
                        rospy.sleep(0.02)

                    goal = LocalPlannerGoal()
                    # goal.intermediate_waypoint =
                    goal.timeout = self.default_timeout
                    self.local_path_planner_action_server(goal=goal)
                    while self._local_path_planner_action_client.get_state() != GoalStatus.ACTIVE:
                        rospy.sleep(0.02)
                else:
                    # reset done above
                    rospy.logwarn('Wrong position, go_to position with second last flag == {} is not a valid value'.format(go_to_position_target[-2]))
                    self._go_to_position_action_client.cancel_all_goals()
                    # stop publishing final_goal_frame tf, stop global_path_planner and local_path_planner action servers
                    self._global_path_planner_action_client.cancel_all_goals()
                    self._local_path_planner_action_client.cancel_all_goals()
        else:
            # reset done above
            rospy.logwarn('Wrong position, go_to position with last flag == {} is not a valid value'.format(go_to_position_target[-1]))
            self._go_to_position_action_client.cancel_all_goals()
            # stop publishing final_goal_frame tf, stop global_path_planner and local_path_planner action servers
            self._global_path_planner_action_client.cancel_all_goals()
            self._local_path_planner_action_client.cancel_all_goals()
        
        self.setting_goals = False


    def _go_to_body_callback(self, data):
        self.reset_go_to_data()
        self.go_to_body_data = data
        go_to_body_target = int(data.id[1:])
        self.setting_goals = True
        if go_to_body_target == -1:
            self.goto_target_id = -1
            # stop goal_finder, global_path_planner, local_path_planner action servers
            self._goal_finder_action_client.cancel_all_goals()
            self._global_path_planner_action_client.cancel_all_goals()
            self._local_path_planner_action_client.cancel_all_goals()
        else:
            self.goto_target_id = go_to_body_target
            self.go_to_body = True
            goal = GoalFinderGoal()
            goal.body_id = data.id
            goal.timeout = self.default_timeout
            self.goal_finder_action_server(goal=goal)
            while self._goal_finder_action_client.get_state() != GoalStatus.ACTIVE:
                rospy.sleep(0.02)
            
            goal = GlobalPlannerGoal()
            # goal.final_goal =
            goal.timeout = self.default_timeout
            self.global_path_planner_action_server(goal=goal)
            while self._global_path_planner_action_client.get_state() != GoalStatus.ACTIVE:
                rospy.sleep(0.02)
            
            goal = LocalPlannerGoal()
            # goal.intermediate_waypoint =
            goal.timeout = self.default_timeout
            self.local_path_planner_action_server(goal=goal)
            while self._local_path_planner_action_client.get_state() != GoalStatus.ACTIVE:
                rospy.sleep(0.02)

        self.setting_goals = False


    def _go_to_group_callback(self, data):
        self.reset_go_to_data()
        self.go_to_group_data = data
        go_to_group_target = int(data.id[3:])
        self.setting_goals = True
        if go_to_group_target == -1:
            self.goto_target_id = -1
            # stop goal_finder, global_path_planner, local_path_planner action servers
            self._goal_finder_action_client.cancel_all_goals()
            self._global_path_planner_action_client.cancel_all_goals()
            self._local_path_planner_action_client.cancel_all_goals()
        else:
            self.goto_target_id = go_to_group_target
            self.go_to_group = True
            goal = GoalFinderGoal()
            goal.group_id = data.id
            goal.timeout = self.default_timeout
            self.goal_finder_action_server(goal=goal)
            while self._goal_finder_action_client.get_state() != GoalStatus.ACTIVE:
                rospy.sleep(0.02)
            
            goal = GlobalPlannerGoal()
            # goal.final_goal =
            goal.timeout = self.default_timeout
            self.global_path_planner_action_server(goal=goal)
            while self._global_path_planner_action_client.get_state() != GoalStatus.ACTIVE:
                rospy.sleep(0.02)
            
            goal = LocalPlannerGoal()
            # goal.intermediate_waypoint =
            goal.timeout = self.default_timeout
            self.local_path_planner_action_server(goal=goal)
            while self._local_path_planner_action_client.get_state() != GoalStatus.ACTIVE:
                rospy.sleep(0.02)
        
        self.setting_goals = False


    def _go_to_person_callback(self, data):
        self.reset_go_to_data()
        self.go_to_person_data = data
        go_to_person_target = data.id
        self.setting_goals = True
        if go_to_person_target == 'p-1':
            self.goto_target_id = -1
            # stop goal_finder, global_path_planner, local_path_planner action servers
            self._goal_finder_action_client.cancel_all_goals()
            self._global_path_planner_action_client.cancel_all_goals()
            self._local_path_planner_action_client.cancel_all_goals()
        else:
            self.goto_target_id = 1  # is str so can not have id as int, 1 by default
            self.go_to_person = True
            goal = GoalFinderGoal()
            goal.person_id = data.id
            goal.timeout = self.default_timeout
            self.goal_finder_action_server(goal=goal)
            while self._goal_finder_action_client.get_state() != GoalStatus.ACTIVE:
                rospy.sleep(0.02)
            
            goal = GlobalPlannerGoal()
            # goal.final_goal =
            goal.timeout = self.default_timeout
            self.global_path_planner_action_server(goal=goal)
            while self._global_path_planner_action_client.get_state() != GoalStatus.ACTIVE:
                rospy.sleep(0.02)
            
            goal = LocalPlannerGoal()
            # goal.intermediate_waypoint =
            goal.timeout = self.default_timeout
            self.local_path_planner_action_server(goal=goal)
            while self._local_path_planner_action_client.get_state() != GoalStatus.ACTIVE:
                rospy.sleep(0.02)
        
        self.setting_goals = False

    
    def _look_at_body_callback(self, data):
        self.reset_look_at_data()
        self.look_at_body_data = data
        look_at_body_target = int(self.look_at_body_data.id[1:])
        humans_id_list = self.humans_id.tolist()
        if float(look_at_body_target) in humans_id_list and look_at_body_target != -1:
            humans_id_index = humans_id_list.index(float(look_at_body_target))
            self.x_pan_target = self.humans_pose[humans_id_index, 0]
            self.y_pan_target = self.humans_pose[humans_id_index, 1]
            self.flag_publish_pan_goal = True
            self.pan_target_id = look_at_body_target
            self.look_at_body = True
        elif look_at_body_target == -1:
            # reset done above
            self._look_at_body_action_client.cancel_all_goals()
        else:
            # reset done above
            rospy.logwarn('Wrong id, look_at body ID {} not in the list of tracked bodies'.format(look_at_body_target))
            self._look_at_body_action_client.cancel_all_goals()


    def _look_at_person_callback(self, data):
        self.reset_look_at_data()
        self.look_at_person_data = data
        look_at_person_target = self.look_at_person_data.id
        if look_at_person_target in self.hri_data[2].keys():
            look_at_person_target = int(self.hri_data[2][look_at_person_target][1:])
            humans_id_list = self.humans_id.tolist()
            if float(look_at_person_target) in humans_id_list:
                humans_id_index = humans_id_list.index(float(look_at_person_target))
                self.x_pan_target = self.humans_pose[humans_id_index, 0]
                self.y_pan_target = self.humans_pose[humans_id_index, 1]
                self.flag_publish_pan_goal = True
                self.pan_target_id = look_at_person_target
                self.look_at_person = True
            else:
                # reset done above
                rospy.logwarn('Wrong id, look_at person ID {} not in the list of tracked persons'.format(look_at_person_target))
                self._look_at_person_action_client.cancel_all_goals()
        elif look_at_person_target == 'p-1':
            # reset done above
            self._look_at_person_action_client.cancel_all_goals()
        else:
            # reset done above
            rospy.logwarn('Wrong id, look_at person ID {} not in the list of tracked persons'.format(look_at_person_target))
            self._look_at_person_action_client.cancel_all_goals()


    def _look_at_group_callback(self, data):
        self.reset_look_at_data()
        self.look_at_group_data = data
        look_at_group_target = int(self.look_at_group_data.id[3:])
        groups_id_list = self.groups_id.tolist()
        if float(look_at_group_target) in groups_id_list and look_at_group_target != -1:
            self.flag_publish_pan_goal = True
            self.pan_target_id = look_at_group_target
            groups_id_index = groups_id_list.index(float(look_at_group_target))
            self.x_pan_target = self.groups_pose[groups_id_index, 0]
            self.y_pan_target = self.groups_pose[groups_id_index, 1]
            self.look_at_group = True
        elif look_at_group_target == -1:
            # reset done above
            self._look_at_group_action_client.cancel_all_goals()
        else:
            # reset done above
            rospy.logwarn('Wrong id, look_at group ID {} not in the list of tracked groups'.format(look_at_group_target))
            self._look_at_group_action_client.cancel_all_goals()


    def _look_at_position_callback(self, data):
        self.reset_look_at_data()
        self.look_at_position_data = data
        look_at_position_target = np.asarray(self.look_at_position_data.look_at_position)
        if look_at_position_target[-1] == -1:
            # reset done above
            self._look_at_position_action_client.cancel_all_goals()
        elif look_at_position_target[-1] == 1:
                self.flag_publish_pan_goal = True
                self.x_pan_target = look_at_position_target[0]
                self.y_pan_target = look_at_position_target[1]
                self.pan_target_id = 1  # is str[] so can not have id as int, 1 by default
                self.look_at_position = True
                if look_at_position_target[-2] == 1:
                    # local frame
                    self.x_pan_target, self.y_pan_target = local_to_global(
                        robot_position=np.array([self.robot_x_position, self.robot_y_position, self.robot_yaw]), 
                        x=np.array([self.x_pan_target, self.y_pan_target])
                    )
                elif look_at_position_target[-2] == 0:
                    # global frame
                    pass
                else:
                    # reset done above
                    rospy.logwarn('Wrong position, look_at position with second last flag == {} is not a valid value'.format(look_at_position_target[-2]))
                    self._look_at_position_action_client.cancel_all_goals()
        else:
            # reset done above
            rospy.logwarn('Wrong position, look_at position with last flag == {} is not a valid value'.format(look_at_position_target[-1]))
            self._look_at_position_action_client.cancel_all_goals()


    def _joint_states_callback(self, data):
        self.joint_states_data = data


    def read_robot_config(self, filename=None):
        if filename is None:
            filename = pkg_resources.resource_filename(
                'sim2d', 'config/robot.yaml')
            rospy.logdebug("No filename provided for the robot configuration, basic robot config loaded")
        elif os.path.isfile(filename):
            rospy.logdebug("Desired robot config loaded")
        else:
            filename = pkg_resources.resource_filename(
                'sim2d', 'config/robot.yaml')
            rospy.logdebug("Desired filename for the robot configuration does not exist, basic robot config loaded")
        config = yaml.load(open(filename), Loader=yaml.FullLoader)
        self.robot_config = RobotConfig(config)


    def read_config(self, filename=None):
        if filename is None:
            filename = pkg_resources.resource_filename(
                'social_mpc', 'config/social_mpc.yaml')
        elif os.path.isfile(filename):
            self.passed_config_loaded = True
        else:
            filename = pkg_resources.resource_filename(
                'social_mpc', 'config/social_mpc.yaml')
        config = yaml.load(open(filename), Loader=yaml.FullLoader)
        self.controller_config = ControllerConfig(config)

        self.goal_finder_enabled = self.controller_config.goal_finder_enabled
        self.path_planner_enabled = self.controller_config.path_planner_enabled
        self.update_goals_enabled = self.controller_config.update_goals_enabled


    def _check_joint_pos_limits(self, p, t):
        if p < self.robot_config.min_pan_angle:
            p = self.robot_config.min_pan_angle
            rospy.logwarn("The pan angle desired is lower than the miminum allowed. The minimum oan angle is set.")
        if p > self.robot_config.max_pan_angle:
            p = self.robot_config.max_pan_angle
            rospy.logwarn("The pan angle desired is greater than the maximum allowed. The maximum oan angle is set.")
        if t < self.robot_config.min_tilt_angle:
            t = self.robot_config.min_tilt_angle
            rospy.logwarn("The tilt angle desired is lower than the miminum allowed. The minimum oan angle is set.")
        if t > self.robot_config.max_tilt_angle:
            t = self.robot_config.max_tilt_angle
            rospy.logwarn("The tilt angle desired is higher than the maximum allowed. The maximum oan angle is set.")
        return p, t


    def _check_all_sensors_ready(self):
        rospy.logdebug("START ALL SENSORS READY")
        # self._check_rtabmap_ready()
        self._check_joint_states_ready()
        rospy.logdebug("ALL SENSORS READY")


    def _check_joint_states_ready(self):
        self.joint_states_data = None
        rospy.logdebug("Waiting for /joint_states to be READY...")
        while self.joint_states_data is None and not rospy.is_shutdown():
            try:
                self.joint_states_data = rospy.wait_for_message("/joint_states", JointState, timeout=5.0)
                rospy.logdebug("Current /joint_states READY=>")

            except:
                rospy.logerr("Current /joint_states not ready yet, retrying for getting joint states")
        if not self.get_indexes:
            # joint_states : head_1_joint, head_2_joint, wheel_left_joint, wheel_right_joint, caster_left_joint, caster_right_joint
            for index, name in enumerate(self.joint_states_data.name):
                if name == "head_1_joint":
                    self.pan_ind = index
                if name == "head_2_joint":
                    self.tilt_ind = index
            self.get_indexes = True

        return self.joint_states_data


    def _check_publishers_connection(self):
        """
        Checks that all the publishers are working
        :return:
        """
        rate = rospy.Rate(10)  # 10hz
        while self._pan_vel_pub.get_num_connections() == 0 and not rospy.is_shutdown():
            rospy.logdebug("No susbribers to _pan_vel_pub yet so we wait and try again")
            try:
                rate.sleep()
            except rospy.ROSInterruptException:
                # This is to avoid error when world is rested, time when backwards.
                pass
        rospy.logdebug("_pan_vel_pub Publisher Connected")
        rospy.logdebug("All Publishers READY")


    def shutdown(self):
        rospy.loginfo("Stopping the PlanningManagerNode")
        self.close()
        rospy.loginfo("Killing the PlanningManagerNode node")
    

    def close(self):
        if self._subscribers:
            for subscriber in self._subscribers:
                subscriber.unregister()

        if self._publishers:
            for publisher in self._publishers:
                if isinstance(publisher, dict):
                    for pub in publisher.values():
                        pub.unregister()
                else:
                    publisher.unregister()
        if self._timers:
            for timer in self._timers:
                timer.shutdown()

        if self._action_server:
            self._action_server.shutdown()
        
        if self._action_client:
            for action_client in self._action_client:
                action_client.cancel_all_goals()
        
        self.ros_4_hri_interface.close()
