#!/usr/bin/env python3
import numpy as np
import tf
import rospy
from nav_msgs.msg import OccupancyGrid, MapMetaData
from robot_behavior.utils.utils import constraint_angle
import actionlib
from spring_msgs.msg import LocalPlannerAction, LocalPlannerFeedback, LocalPlannerResult
from geometry_msgs.msg import Twist, Vector3
from std_msgs.msg import Float32
from robot_behavior.utils.drl_local_path_planner import LocalPathPlannerDRL
from robot_behavior.utils.utils_drl_local_path_planner import *
from human_aware_navigation_rl.agent.sac.sac_base import SACBase
import exputils as eu

class RLLocalPathPlannerActionServer:
    def __init__(self, name):
        self._action_name = name

        self.map_frame = rospy.get_param('~map_frame', 'map')
        self.robot_frame = rospy.get_param('~robot_frame', 'base_footprint')
        self.local_occupancy_map_topic = rospy.get_param('~local_occupancy_map_topic_rl', '/slam/local_map_rl_navigation')
        self.local_social_cost_map_topic = rospy.get_param('~local_social_cost_map_topic', '/slam/local_social_cost_map')
        self.timer_error_max_time = rospy.get_param('~timer_error_max_time', 5)  # seconds
        self.final_goal_frame = rospy.get_param('~final_goal_frame', 'final_goal')
        self.intermediate_waypoint_goal_frame = rospy.get_param('~intermediate_waypoint_goal_frame', 'intermediate_waypoint_goal')
        self.cmd_vel_topic = rospy.get_param('~cmd_vel_topic', '/nav_vel')
        self.namespace_slam = rospy.get_param('~namespace_slam', '/rtabmap')
        self.rl_publishing_rate = rospy.get_param('~rl_publishing_rate', 10)
        self.time_inference_rl_topic = rospy.get_param('~time_inference_rl_topic', '/time_inference_rl')


        self._feedback = LocalPlannerFeedback()
        self._result = LocalPlannerResult()
        
        self.local_map_data = None
        self.local_social_cost_map_data = None
        self.rtabmap_ready = False
        self.local_map_static = None
        self.local_social_cost_map = None

        self.x_wp = 0.
        self.y_wp = 0.
        self.yaw_wp = 0.
        self.robot_x_position = 0.
        self.robot_y_position = 0.
        self.robot_yaw = 0.
        self.robot_position = [0, 0]

        self.intermediate_waypoint = None

        self.tf_robot_pose_error = False
        self.tf_intermediate_wp_pose_error = False
        self.flag_actions_okay = False

        # Trained with obstacle
        # folder_path = '/home/ros/robot_behavior_ws/modules/human_aware_navigation_rl/human_aware_navigation_rl/trained_models/campaign_real_env_obstacles/experiments/'
        # import_weight = eu.AttrDict(
        #    navigation_encoder_actor = folder_path+'experiment_000010/repetition_000000/checkpoints/final_nav_encoder_actor_state_dict.pt',
        #    actor = folder_path+'experiment_000010/repetition_000000/checkpoints/final_actor_state_dict.pt',
        #    critic = folder_path+'experiment_000010/repetition_000000/checkpoints/final_critic_state_dict.pt'
        # )

        # Trained without obstacle
        # folder_path = '/home/ros/robot_behavior_ws/modules/human_aware_navigation_rl/human_aware_navigation_rl/trained_models/campaign_real_env_warning_zone_bis/experiments/'
        # import_weight = eu.AttrDict(
        #    navigation_encoder_actor = folder_path+'experiment_000013/repetition_000000/checkpoints/final_nav_encoder_actor_state_dict.pt',
        #    actor = folder_path+'experiment_000013/repetition_000000/checkpoints/final_actor_state_dict.pt',
        #    critic = folder_path+'experiment_000013/repetition_000000/checkpoints/final_critic_state_dict.pt'
        # )

        # config_agent = eu.AttrDict(
        #     list_of_input_to_memorize = [['static_occupancy_grid'],['angle_to_goal_position', 'distance_to_goal_position']],
        #     list_of_input_to_encode = [['static_occupancy_grid'],['angle_to_goal_position', 'distance_to_goal_position']],
        #     latent_shape = 288, 
        #     normalize_latent_space = True,
        # )

        # Trained with static humans
        folder_path = '/home/ros/robot_behavior_ws/modules/human_aware_navigation_rl/human_aware_navigation_rl/trained_models/campaign_real_env_human_collision_penalty_optim/experiments/'
        import_weight = eu.AttrDict(
           navigation_encoder_actor = folder_path+'experiment_000015/repetition_000000/checkpoints/final_nav_encoder_state_dict.pt',
           actor = folder_path+'experiment_000015/repetition_000000/checkpoints/final_actor_state_dict.pt',
           critic = folder_path+'experiment_000015/repetition_000000/checkpoints/final_critic_state_dict.pt'
        )

        config_agent = eu.AttrDict(
            list_of_input_to_memorize = [['static_occupancy_grid','dynamic_social_grid'],['angle_to_goal_position', 'distance_to_goal_position']],
            list_of_input_to_encode = [['static_occupancy_grid','dynamic_social_grid'],['angle_to_goal_position', 'distance_to_goal_position']],
            latent_shape = 288, 
            normalize_latent_space = True,
        )

        config_env  = eu.AttrDict(
            raycast_field_of_view = 360,
            raycast_angle_between_two_ray=10,
            vmin = 0,
            vmax = 0.5,
            wmin = -0.5,
            wmax = 0.5,
            grid_shape=(1,60,60),
            local_map_depth = 3

        )
        class_agent_name = rospy.get_param('~agent_class', 'SACBase')
        # Creation of class attribute
        class_agent = eval(class_agent_name)
        self.agent = class_agent(env=config_env, import_weight = import_weight, config=config_agent, device = "cpu")
        self.rl_path_planner = LocalPathPlannerDRL(list_of_input_to_encode = self.agent.unsorted_list_of_input, config_env=config_env)
        self.goal_position = [0., 0.]
        self.observation=None
        self.current_time = 0
        self.time_taken_for_inference = 0
        
        self.lin_vel, self.ang_vel = 0, 0

        self.init_ros_subscriber_and_publicher()

        self.tf_broadcaster = tf.TransformBroadcaster()  # Publish Transform to look for human
        self.tf_listener = tf.TransformListener()  # Listen Transform to look for human

        self._check_all_sensors_ready()

        rospy.on_shutdown(self.shutdown)

        self._a_server = actionlib.SimpleActionServer(self._action_name, LocalPlannerAction, execute_cb=self.execute_cb, auto_start=False)
        self._a_server.start()

        rospy.loginfo("{} Initialization Ended".format(self._action_name))


    def execute_cb(self, goal):
        # helper variables
        rate = rospy.Rate(self.rl_publishing_rate)
        self.init_time = rospy.Time.now()
        self.current_time = (rospy.Time.now() - self.init_time).to_sec()
        self.intermediate_waypoint = None

        # publish info to the console for the user
        rospy.loginfo('Goal message : {}'.format(goal))
        self.intermediate_waypoint = goal.intermediate_waypoint

        # At time step 0, intermediate_waypoint_goal_frame = robot position
        self.flag_actions_okay = False
        self.tf_robot_pose_error = False
        self.tf_intermediate_wp_pose_error = False
        self.get_robot_pose()
        self.timer_tf_intermediate_wp_pose_error = None
        self.timer_tf_robot_pose_error = None

        # start executing the action
        while not rospy.is_shutdown():
            if self._a_server.is_preempt_requested():
                self._a_server.set_preempted()
                break

            self.current_time = (rospy.Time.now() - self.init_time).to_sec()
            rospy.logdebug('self.current_time: {}'.format(self.current_time))
            # create messages that are used to publish feedback/result
            self._feedback = LocalPlannerFeedback()
            self._result = LocalPlannerResult()

            # publish the feedback
            self._publish_feedback(goal)
            rate.sleep()

            if self.current_time > goal.timeout:
                text = "Timeout"
                rospy.loginfo('{}: Aborted. {}'.format(self._action_name, text))
                self._a_server.set_aborted(self._result)
                break
            
            if self.tf_robot_pose_error:
                if not self.timer_tf_robot_pose_error:
                    self.timer_tf_robot_pose_error = rospy.Time.now()
                error_time = (rospy.Time.now() - self.timer_tf_robot_pose_error).to_sec()
                rospy.logwarn("Could not get transformation between {} and {}, no transformation available for {} seconds".format(self.map_frame, self.robot_frame, error_time))
                if error_time > self.timer_error_max_time:
                    text = "Could not get transformation between {} and {}, no transformation available during more than {} seconds. So the {} will be aborted".format(self.map_frame, self.robot_frame, self.timer_error_max_time, self._action_name)
                    rospy.loginfo('{}: Aborted. {}'.format(self._action_name, text))
                    self._a_server.set_aborted(self._result)
                    break

            if self.tf_intermediate_wp_pose_error:
                if not self.timer_tf_intermediate_wp_pose_error:
                    self.timer_tf_intermediate_wp_pose_error = rospy.Time.now()
                error_time = (rospy.Time.now() - self.timer_tf_intermediate_wp_pose_error).to_sec()
                rospy.logwarn("Could not get transformation between {} and {}, no transformation available for {} seconds".format(self.map_frame, self.intermediate_waypoint_goal_frame, error_time))
                if error_time > self.timer_error_max_time:
                    text = "Could not get transformation between {} and {}, no transformation available during more than {} seconds. So the {} will be aborted".format(self.map_frame, self.intermediate_waypoint_goal_frame, self.timer_error_max_time, self._action_name)
                    rospy.loginfo('{}: Aborted. {}'.format(self._action_name, text))
                    self._a_server.set_aborted(self._result)
                    break
            
            self.flag_actions_okay = True
            self.step(self.current_time)

        self.flag_actions_okay = False
        self._result_msg(goal)


    def _result_msg(self, goal):
        self.current_time = (rospy.Time.now() - self.init_time).to_sec()
        self._result.final_time = self.current_time


    def _publish_feedback(self, goal):
        self._feedback.current_time = self.current_time
        if self.flag_actions_okay:
            self._feedback.robot_pose.position.x = self.robot_x_position
            self._feedback.robot_pose.position.y = self.robot_x_position
            self._feedback.robot_pose.orientation.x = self.quat_robot[0]
            self._feedback.robot_pose.orientation.y = self.quat_robot[1]
            self._feedback.robot_pose.orientation.z = self.quat_robot[2]
            self._feedback.robot_pose.orientation.w = self.quat_robot[3]
        self._a_server.publish_feedback(self._feedback)


    def step(self, init_time):
        r''' Step the ros wrapper'''

        # Localizing the robot in the environment
        if not self.get_robot_pose():
            self.tf_robot_pose_error = True
        else:
            self.tf_robot_pose_error = False
            self.timer_tf_robot_pose_error = None
        
        # Get the intermediate_waypoint_goal_frame tf
        if not self.get_waypoint_pose():
            self.tf_intermediate_wp_pose_error = True
        else:
            self.tf_intermediate_wp_pose_error = False
            self.timer_tf_intermediate_wp_pose_error = None

        rospy.logdebug('tf_intermediate_wp_pose_error: {}'.format(self.tf_intermediate_wp_pose_error))
        rospy.logdebug('tf_robot_pose_error'.format(self.tf_robot_pose_error))

        if self.goal_position:
            if self.observation is not None:
                self.observation.linear_velocity = self.lin_vel
                self.observation.angular_velocity = self.ang_vel
                self.current_time = rospy.Time.now()
                action, prediction = self.rl_path_planner.predict(agent = self.agent, observation = self.observation)
                self.time_taken_for_inference = (rospy.Time.now() - self.current_time).to_sec()
                self._time_inference_pub.publish(Float32(self.time_taken_for_inference))

                self.lin_vel = action[0]
                self.ang_vel = action[1]
            else:
                self.lin_vel=0
                self.ang_vel=0


    def init_ros_subscriber_and_publicher(self):
        r''' Initialize the subscribers and publishers '''
        self._subscribers = []
        self._publishers = []
        self._action_server = []
        self._timers = []

        ### ROS Subscribers
        self._subscribers.append(rospy.Subscriber(self.local_occupancy_map_topic, OccupancyGrid, callback=self._local_map_callback, queue_size=1))
        self._subscribers.append(rospy.Subscriber(self.local_social_cost_map_topic, OccupancyGrid, callback=self._local_social_cost_map_callback, queue_size=1))

        ### ROS Publishers
        self._cmd_vel_pub = rospy.Publisher(self.cmd_vel_topic, Twist, queue_size=1)
        self._publishers.append(self._cmd_vel_pub)
        self._local_map_pub = rospy.Publisher(self.namespace_slam + '/occupancy_grid_for_drl_nav', OccupancyGrid, queue_size=10, latch=True)
        self._local_map_raycast_pub = rospy.Publisher(self.namespace_slam + '/raycast_for_drl_nav', OccupancyGrid, queue_size=10, latch=True)
        self._local_map_closest_obstacle_pub = rospy.Publisher(self.namespace_slam + '/closest_obstacle_map_for_drl_nav', OccupancyGrid, queue_size=10, latch=True)
        self._publishers.append(self._local_map_pub)
        self._publishers.append(self._local_map_raycast_pub)
        self._publishers.append(self._local_map_closest_obstacle_pub)

        self._time_inference_pub = rospy.Publisher(self.time_inference_rl_topic, Float32, queue_size=1)
        self._publishers.append(self._time_inference_pub)

        self._timers.append(rospy.Timer(rospy.Duration(1/self.rl_publishing_rate), callback=self._pub_vel_callback))

        rospy.loginfo("Initializing the %s " % self._action_name)

        self._check_publishers_connection()


    def publish_tf_go_to_goals(self):
        current_time = rospy.Time.now()
        # to debug 
        # quat_final = tf.transformations.quaternion_from_euler(0, 0, self.yaw_final)
        # self.x_final, self.y_final = self.goal_position
        # self.tf_broadcaster.sendTransform(
        #     (self.x_final, self.y_final, 0),
        #     quat_final,
        #     current_time,
        #     self.final_goal_frame,
        #     self.map_frame
        # )
        if self.flag_waypoint_okay:
            self.tf_broadcaster.sendTransform(
                (self.x_wp, self.y_wp, 0),
                self.quat_wp,
                current_time,
                self.intermediate_waypoint_goal_frame,
                self.map_frame
            )


    def _pub_vel_callback(self, event):
        r''' Function that checks the velocities in the memory and send it to the robot
         
          Warning : This functions locks the reading of the velocitites while sending it to the robot '''
        if self.flag_actions_okay:
            self.lin_vel, self.ang_vel = self._check_joint_vel_limits(self.lin_vel, self.ang_vel)
            rospy.logdebug('lin_vel: {} \t ang_vel: {}'.format(self.lin_vel, self.ang_vel))
            twist_msg = Twist(Vector3(self.lin_vel, 0, 0), Vector3(0, 0, self.ang_vel))
            self._cmd_vel_pub.publish(twist_msg)


    def get_waypoint_pose(self):
        try:
            map_2_waypoint = self.tf_listener.lookupTransform(self.map_frame, self.intermediate_waypoint_goal_frame, rospy.Time(0))
            #rospy.logwarn("Could not get transformation between {} and {} at correct time. Using latest available.".format(self.map_frame, self.robot_frame))
        except:
            rospy.logerr("Could not get transformation between {} and {}.".format(self.map_frame, self.intermediate_waypoint_goal_frame))
            return False
        self.x_wp, self.y_wp, _ = map_2_waypoint[0]
        (_, _, yaw) = tf.transformations.euler_from_quaternion([
            map_2_waypoint[1][0],
            map_2_waypoint[1][1],
            map_2_waypoint[1][2],
            map_2_waypoint[1][3]
        ])
        self.yaw_wp = float(constraint_angle(yaw))
        self.goal_position = [self.x_wp, self.y_wp]
        return True


    def get_robot_pose(self):
        r""" Function that compute the position of the robot on the map """
        # try:
        #     map_2_robot = self.tf_listener.lookupTransform(self.map_frame, self.robot_frame, self.timestamp_tracking_latest)
        # except:
        try:
            map_2_robot = self.tf_listener.lookupTransform(self.map_frame, self.robot_frame, rospy.Time(0))
            #rospy.logwarn("Could not get transformation between {} and {} at correct time. Using latest available.".format(self.map_frame, self.robot_frame))
        except:
            rospy.logerr("Could not get transformation between {} and {}.".format(self.map_frame, self.robot_frame))
            return False
        self.robot_x_position, self.robot_y_position, _ = map_2_robot[0]
        (_, _, yaw) = tf.transformations.euler_from_quaternion([
            map_2_robot[1][0],
            map_2_robot[1][1],
            map_2_robot[1][2],
            map_2_robot[1][3]
        ])
        self.robot_yaw = float(constraint_angle(yaw))
        self.robot_position = np.array([self.robot_x_position, self.robot_y_position])
        self.quat_robot = tf.transformations.quaternion_from_euler(0, 0, self.robot_yaw)
        return True


    def _local_map_callback(self, data):
        self.local_map_data = data
        self.x_local_map = self.local_map_data.info.origin.position.x
        self.y_local_map = self.local_map_data.info.origin.position.y
        self.local_map_width = self.local_map_data.info.width
        self.local_map_height = self.local_map_data.info.height
        self.local_map_resolution = self.local_map_data.info.resolution
        self.local_map_size = [[self.x_local_map, self.x_local_map + self.local_map_width*self.local_map_resolution],[self.y_local_map, self.y_local_map + self.local_map_height*self.local_map_resolution]]
        self.last_shape_local_map = (self.local_map_height, self.local_map_width)
        self.local_static_map = (np.asarray(self.local_map_data.data) / 100).reshape(self.last_shape_local_map)
        flag_updated_robot_pose = self.get_robot_pose()
            
        if flag_updated_robot_pose and self.local_social_cost_map is not None:
            self.observation, self.raycast_map, self.closest_obstacle_map = self.rl_path_planner.compute_observation_from_data(
                                                                                                            static_occupancy_grid=np.flip(self.local_static_map, axis=0),
                                                                                                            dynamic_social_grid =np.flip(self.local_social_cost_map, axis=0),
                                                                                                            robot_position=self.robot_position,
                                                                                                            robot_orientation=self.robot_yaw,
                                                                                                            goal_position=np.array(self.goal_position),
                                                                                                            linear_velocity=self.lin_vel,
                                                                                                            angular_velocity=self.ang_vel,
                                                                                                            goal_yaw=self.yaw_wp
            )
            if self.raycast_map is not None :
                self.publish_raycast_map(self.raycast_map)
            if self.closest_obstacle_map is not None :
                self.publish_closest_obstacle_map(self.closest_obstacle_map)
            if self.observation.static_occupancy_grid is not None :
                #self.publish_local_map(np.flip(self.observation.static_occupancy_grid, axis=1))
                self.publish_local_map(np.flip(np.clip(self.observation.static_occupancy_grid + self.observation.dynamic_social_grid,0,1), axis=1)) 


    def publish_raycast_map(self, map):
        r''' Publish raycast map '''
        self._local_map_raycast_pub.publish(self.set_occupancy_grid_object(map))


    def publish_closest_obstacle_map(self, map):
        r''' Publish closest obstacle map '''
        self._local_map_closest_obstacle_pub.publish(self.set_occupancy_grid_object(map))


    def publish_local_map(self, map):
        r''' Publish local occupancy grid map '''
        self._local_map_pub.publish(self.set_occupancy_grid_object(map))


    def set_occupancy_grid_object(self, map):
        r''' Set the occupancy grid map object from the map'''
        occupancy_grid = OccupancyGrid()
        occupancy_grid.info = MapMetaData()
        occupancy_grid.info.map_load_time = rospy.Time.now()
        resolution = self.local_map_data.info.resolution
        if self.local_map_width != np.shape(map)[-1] :
            resolution = self.local_map_data.info.resolution * self.local_map_width/np.shape(map)[-1]
        occupancy_grid.info.resolution =  resolution
        occupancy_grid.info.width = np.shape(map)[1]
        occupancy_grid.info.height = np.shape(map)[1]
        occupancy_grid.info.origin.position.x = self.local_map_data.info.origin.position.x
        occupancy_grid.info.origin.position.y = self.local_map_data.info.origin.position.y
        occupancy_grid.data = (map*100).flatten().astype(np.int8)
        occupancy_grid.header = self.local_map_data.header
        return occupancy_grid

    def _local_social_cost_map_callback(self, data):
        self.local_social_cost_map_data = data
        self.x_local_social_cost_map = self.local_social_cost_map_data.info.origin.position.x
        self.y_local_social_cost_map = self.local_social_cost_map_data.info.origin.position.y
        self.local_social_cost_map_width = self.local_social_cost_map_data.info.width
        self.local_social_cost_map_height = self.local_social_cost_map_data.info.height
        self.local_social_cost_map_resolution = self.local_social_cost_map_data.info.resolution
        self.local_social_cost_map_size = [[self.x_local_social_cost_map, self.x_local_social_cost_map + self.local_social_cost_map_width*self.local_social_cost_map_resolution],[self.y_local_social_cost_map, self.y_local_social_cost_map + self.local_social_cost_map_height*self.local_social_cost_map_resolution]]
        self.last_shape_local_social_cost_map = (self.local_social_cost_map_height, self.local_social_cost_map_width)
        self.local_social_cost_map = (np.asarray(self.local_social_cost_map_data.data) / 100).reshape(self.last_shape_local_social_cost_map)
        self.local_social_cost_map = cv2.resize(self.local_social_cost_map, (60,60), interpolation = cv2.INTER_NEAREST)

    def _check_joint_vel_limits(self, lin_vel, ang_vel):
        flag_high_forward_speed = False
        if lin_vel > self.rl_path_planner.forward_velocity_limit[1]:
            rospy.logwarn("The linear velocity is greater than the maximum allowed. The maximum velocity is set.")
            lin_vel = self.rl_path_planner.forward_velocity_limit[1]
            flag_high_forward_speed = True
        if lin_vel < self.rl_path_planner.forward_velocity_limit[0]:
            rospy.logwarn("The linear velocity is lower than the minimum allowed. The minimum velocity is set.")
            lin_vel = self.rl_path_planner.forward_velocity_limit[0]
        if ang_vel > self.rl_path_planner.angular_velocity_limit[1] and flag_high_forward_speed:
            rospy.logwarn("The angular velocity is greater than the maximum allowed. The maximum velocity is set.")
            ang_vel =  self.rl_path_planner.angular_velocity_limit[1]
        if ang_vel < self.rl_path_planner.angular_velocity_limit[0] and flag_high_forward_speed:
            rospy.logwarn("The angular velocity is lower than the minimum allowed. The minimum velocity is set.")
            ang_vel = self.rl_path_planner.angular_velocity_limit[0]
        return lin_vel, ang_vel
    
    def _check_publishers_connection(self):
        """
        Checks that all the publishers are working
        :return:
        """
        rate = rospy.Rate(10)  # 10hz
        while self._cmd_vel_pub.get_num_connections() == 0 and not rospy.is_shutdown():
            rospy.logdebug("No susbribers to _cmd_vel_pub yet so we wait and try again")
            try:
                rate.sleep()
            except rospy.ROSInterruptException:
                # This is to avoid error when world is rested, time when backwards.
                pass
        rospy.logdebug("_cmd_vel_pub Publisher Connected")


    def _check_all_sensors_ready(self):
        rospy.logdebug("START ALL SENSORS READY")
        self._check_rtabmap_ready()
        self._check_local_map_ready()
        self._check_local_social_cost_map_ready()
        rospy.logdebug("ALL SENSORS READY")


    def _check_local_map_ready(self):
        self.local_map_data = None
        rospy.logdebug("Waiting for {} to be READY...".format(self.local_occupancy_map_topic))
        while self.local_map_data is None and not rospy.is_shutdown():
            try:
                self.local_map_data = rospy.wait_for_message(self.local_occupancy_map_topic, OccupancyGrid, timeout=5.0)
                rospy.logdebug("Current {} READY=>".format(self.local_occupancy_map_topic))

            except:
                rospy.logerr("Current {} not ready yet, retrying for getting local map".format(self.local_occupancy_map_topic))
        return self.local_map_data


    def _check_local_social_cost_map_ready(self):
        self.local_social_cost_map_data = None
        rospy.logdebug("Waiting for {} to be READY...".format(self.local_social_cost_map_topic))
        while self.local_social_cost_map_data is None and not rospy.is_shutdown():
            try:
                self.local_social_cost_map_data = rospy.wait_for_message(self.local_social_cost_map_topic, OccupancyGrid, timeout=5.0)
                rospy.logdebug("Current {} READY=>".format(self.local_social_cost_map_topic))

            except:
                rospy.logerr("Current {} not ready yet, retrying for getting local map".format(self.local_social_cost_map_topic))
        return self.local_social_cost_map_data


    def _check_rtabmap_ready(self):
        rospy.logdebug("Waiting for rtabmap pose to be READY...")
        while self.rtabmap_ready is None and not rospy.is_shutdown():
            try:
                self.tf_listener.waitForTransform(self.map_frame, self.robot_frame, rospy.Time(0), rospy.Duration(5.0))
                self.rtabmap_ready = True
                rospy.logdebug("Current rtabmap pose READY=>")

            except:
                rospy.logerr("Current rtabmap pose not ready yet, retrying for getting rtabmap pose")
        return self.rtabmap_ready


    def shutdown(self):
        rospy.loginfo("Stopping the %s " % self._action_name)
        self.close()
        rospy.loginfo("Killing the %s node" % self._action_name)


    def close(self):
        if self._subscribers:
            for subscriber in self._subscribers:
                subscriber.unregister()

        if self._publishers:
            for publisher in self._publishers:
                if isinstance(publisher, dict):
                    for pub in publisher.values():
                        pub.unregister()
                else:
                    publisher.unregister()
        if self._timers:
            for timer in self._timers:
                timer.shutdown()

        if self._action_server:
            self._action_server.shutdown()
