#!/usr/bin/env python3
import numpy as np
import tf
import sys
import rospy
from nav_msgs.msg import OccupancyGrid
import time
from robot_behavior.utils.ros_4_hri_interface import ROS4HRIInterface
from robot_behavior.utils.social_spaces import SocialSpaces, Group
from robot_behavior.utils.utils import constraint_angle, global_to_local


class SocialCostMapNode:
    def __init__(self):
        self.local_occupancy_map_topic = rospy.get_param('~local_occupancy_map_topic', '/slam/local_occupancy_map')
        self.local_social_cost_map_topic = rospy.get_param('~local_social_cost_map_topic', '/slam/local_social_cost_map')
        self.max_humans_world = rospy.get_param('max_humans_world', 20)
        self.max_groups_world = rospy.get_param('max_groups_world', 10)
        self.map_frame = rospy.get_param('~map_frame', 'map')
        self.robot_frame = rospy.get_param('~robot_frame', 'base_footprint')

        self.local_map_data = None
        self.rtabmap_ready = False
        self.ssn = None
        self.init_done = False

        self.init_ros_subscriber_and_publicher()

        self.tf_listener = tf.TransformListener()  # Listen Transform to look for human

        self._check_all_sensors_ready()

        self.init()

        rospy.on_shutdown(self.shutdown)

        rospy.loginfo("SocialCostMapNode Initialization Ended")


    def init(self):
        self.ros_4_hri_interface = ROS4HRIInterface()
        self.hri_data = None
        self.reset_state()        

        if self.ssn is None:
            local_map_size = [
                [self.local_map_data.info.origin.position.x, self.local_map_data.info.origin.position.x + self.local_map_data.info.width*self.local_map_data.info.resolution],
                [self.local_map_data.info.origin.position.y, self.local_map_data.info.origin.position.y + self.local_map_data.info.height*self.local_map_data.info.resolution]]
            self.ssn = SocialSpaces(
                    bbox=local_map_size,
                    map_shape=(self.local_map_data.info.height, self.local_map_data.info.width))
        
        self.init_done = True

    def reset_state(self):
        self.human_id_default = -1
        self.group_id_default = -1

        nb_humans = self.max_humans_world
        self.humans_id = np.ones(nb_humans) * self.human_id_default
        self.humans_group_id = np.ones(nb_humans) * self.group_id_default
        self.groups_id = np.ones(self.max_groups_world) * self.group_id_default
        self.humans_visible = np.ones(nb_humans)
        self.humans_pose = np.zeros((nb_humans, 3))  # x, y, angle
        self.robot_pose = np.zeros(3)  # x, y, angle
        self.humans_velocity = np.zeros((nb_humans, 2))  # linear, angular
        self.groups_pose = np.zeros((self.max_groups_world, 2))  # x, y
        self.groups_visible = np.ones(self.max_groups_world)


    def init_ros_subscriber_and_publicher(self):
        r''' Initialize the subscribers and publishers '''
        self._subscribers = []
        self._publishers = []
        self._action_server = []
        self._timers = []

        ### ROS Subscribers
        self._subscribers.append(rospy.Subscriber(self.local_occupancy_map_topic, OccupancyGrid, callback=self._local_map_callback, queue_size=1))

        ### ROS Publishers
        self._local_social_cost_map_pub = rospy.Publisher(self.local_social_cost_map_topic, OccupancyGrid, queue_size=1)
        self._publishers.append(self._local_social_cost_map_pub)

        rospy.loginfo("Initializing the SocialCostMapNode")

        # self._check_publishers_connection()


    def publish_social_cost_map(self):
        self.reset_state()
        self.hri_data = self.ros_4_hri_interface.get_data()
        self.robot_pose = np.array([self.robot_x_position, self.robot_y_position, self.robot_yaw])
        
        for human_id, human_data in self.hri_data[0].items():
            human_id = int(human_id[1:])
            humans_id_list = self.humans_id.tolist()
            if human_id in humans_id_list:
                index = humans_id_list.index(human_id)
            elif self.human_id_default in humans_id_list:
                index = humans_id_list.index(self.human_id_default)
            else:
                rospy.logerr("No more space in state.humans_ arrays. Number max is {}. It is fixed because of the use of SharedMemory multiprocessing".format(self.max_humans_world))
            self.humans_id[index] = human_id
            self.humans_visible[index] = 0.  # how to get it ?
            if len(human_data) >= 3:
                self.humans_group_id[index] = int(human_data[2][3:])  # 0: transform, 1:person_id, 2:group_id if group

            human_pose = human_data[0].transform.translation
            human_quat = human_data[0].transform.rotation
            x_pos, y_pos = human_pose.x, human_pose.y
            (_, _, yaw) = tf.transformations.euler_from_quaternion([
                human_quat.x,
                human_quat.y,
                human_quat.z,
                human_quat.w])

            lin = 0.  # how to get it ?
            ang = 0.  # how to get it ?

            self.humans_pose[index, :] = [x_pos, y_pos, constraint_angle(yaw)]
            self.humans_velocity[index, :] = [lin, ang]

        # rospy.logdebug('humans_id : {}'.format(self.humans_id.tolist()))

        for group_id, group_data in self.hri_data[1].items():
            group_id  = int(group_id[3:])
            groups_id_list = self.groups_id.tolist()
            if group_id in groups_id_list:
                index = groups_id_list.index(group_id)
            elif self.group_id_default in groups_id_list:
                index = groups_id_list.index(self.group_id_default)
            else:
                rospy.logerr("No more space in state.groups_ arrays. Number max is {}. It is fixed because of the use of SharedMemory multiprocessing".format(self.max_groups_world))
            self.groups_id[index] = group_id
            self.groups_visible[index] = 0.  # how to get it ?

            group_pose = group_data[0].transform.translation
            x_pos, y_pos = group_pose.x, group_pose.y
            self.groups_pose[index, :] = [x_pos, y_pos]

            
        # rospy.logdebug('groups_id ; {}'.format(self.groups_id.tolist()))

        groups = None
        ids = np.where(self.humans_visible == 0)[0]
        n_human = len(ids)
        # (x, y, direction, velocity, is_sitting)
        humans = np.zeros((n_human, 5))
        for j in range(n_human):
            i = ids[j]
            global_position = self.humans_pose[i, :2]
            local_position = global_to_local(self.robot_pose, global_position)
            local_angle = self.humans_pose[i, - 1] - self.robot_pose[-1]
            humans[j, :] = * \
                local_position, local_angle, self.humans_velocity[i, 0], 0
        if n_human > 1:
            groups = []
            for idx, group_id in enumerate(self.groups_id):
                if group_id != -1:
                    center = global_to_local(self.robot_pose, self.groups_pose[idx])
                    person_ids = np.where(self.humans_group_id == group_id)[0]
                    if person_ids.shape[0] > 0:
                        groups.append(Group(center=center, person_ids=person_ids))
                else:
                    break

        f_dsz = self.ssn.calc_dsz(humans, groups=groups)
        occupancy_grid = self.local_map_data
        occupancy_grid.info.map_load_time = rospy.Time.now()
        occupancy_grid.data = (f_dsz * 100).flatten().astype(np.int8)
        self._local_social_cost_map_pub.publish(occupancy_grid)


    def _local_map_callback(self, data):
        self.local_map_data = data
        self.x_local_map = self.local_map_data.info.origin.position.x
        self.y_local_map = self.local_map_data.info.origin.position.y
        self.local_map_width = self.local_map_data.info.width
        self.local_map_height = self.local_map_data.info.height
        self.local_map_resolution = self.local_map_data.info.resolution
        self.local_map_size = [[self.x_local_map, self.x_local_map + self.local_map_width*self.local_map_resolution],[self.y_local_map, self.y_local_map + self.local_map_height*self.local_map_resolution]]
        self.last_shape_local_map = (self.local_map_height, self.local_map_width)
        self.local_static_map = (np.asarray(self.local_map_data.data) / 100).reshape(self.last_shape_local_map)
        if self.init_done:
            if self.get_robot_pose():
                self.publish_social_cost_map()

    
    def get_robot_pose(self):
        r""" Function that compute the position of the robot on the map """
        # try:
        #     map_2_robot = self.tf_listener.lookupTransform(self.map_frame, self.robot_frame, self.timestamp_tracking_latest)
        # except:
        try:
            map_2_robot = self.tf_listener.lookupTransform(self.map_frame, self.robot_frame, rospy.Time(0))
            # rospy.logwarn("Could not get transformation between {} and {} at correct time. Using latest available.".format(self.map_frame, self.robot_frame))
        except:
            rospy.logerr("Could not get transformation between {} and {}.".format(self.map_frame, self.robot_frame))
            return False
        self.robot_x_position, self.robot_y_position, _ = map_2_robot[0]
        (_, _, yaw) = tf.transformations.euler_from_quaternion([map_2_robot[1][0],
                                                                map_2_robot[1][1],
                                                                map_2_robot[1][2],
                                                                map_2_robot[1][3]])
        self.robot_yaw = float(constraint_angle(yaw))
        self.robot_position = np.array([self.robot_x_position, self.robot_y_position])
        return True


    def _check_all_sensors_ready(self):
        rospy.logdebug("START ALL SENSORS READY")
        self._check_rtabmap_ready()
        self._check_local_map_ready()
        rospy.logdebug("ALL SENSORS READY")


    def _check_local_map_ready(self):
        self.local_map_data = None
        rospy.logdebug("Waiting for {} to be READY...".format(self.local_occupancy_map_topic))
        while self.local_map_data is None and not rospy.is_shutdown():
            try:
                self.local_map_data = rospy.wait_for_message(self.local_occupancy_map_topic, OccupancyGrid, timeout=5.0)
                rospy.logdebug("Current {} READY=>".format(self.local_occupancy_map_topic))

            except:
                rospy.logerr("Current {} not ready yet, retrying for getting local map".format(self.local_occupancy_map_topic))
        return self.local_map_data


    def _check_rtabmap_ready(self):
        rospy.logdebug("Waiting for rtabmap pose to be READY...")
        while self.rtabmap_ready is None and not rospy.is_shutdown():
            try:
                self.tf_listener.waitForTransform(self.map_frame, self.robot_frame, rospy.Time(0), rospy.Duration(5.0))
                self.rtabmap_ready = True
                rospy.logdebug("Current rtabmap pose READY=>")

            except:
                rospy.logerr("Current rtabmap pose not ready yet, retrying for getting rtabmap pose")
        return self.rtabmap_ready


    def shutdown(self):
        rospy.loginfo("Stopping the SocialCostMapNode node")
        self.close()
        rospy.loginfo("Killing the SocialCostMapNode node")


    def close(self):
        if self._subscribers:
            for subscriber in self._subscribers:
                subscriber.unregister()

        if self._publishers:
            for publisher in self._publishers:
                if isinstance(publisher, dict):
                    for pub in publisher.values():
                        pub.unregister()
                else:
                    publisher.unregister()
        if self._timers:
            for timer in self._timers:
                timer.shutdown()

        if self._action_server:
            self._action_server.shutdown()

        self.ros_4_hri_interface.close()
