import numpy as np
import exputils as eu
import cv2
from robot_behavior.utils.utils_drl_local_path_planner import *
from human_aware_navigation_rl.agent.sac.sac_base import SACBase
from human_aware_navigation_rl.agent.ddqn.ddqn import DDQN
from human_aware_navigation_rl.env.utils.distance_grid_from_occupancy_grid import raycast_from_occupancy_grid
from human_aware_navigation_rl.env.utils.distance_grid_from_occupancy_grid import distance_and_angle_to_nearest_obstacle_from_occupancy_grid

class LocalPathPlannerDRL:
    r''' LocalPathPlanner based of Deep Reinforcement Learning trained agent '''
    def __init__(self, list_of_input_to_encode, config_env):
        
        # Observation Space
        self.list_of_input_to_encode = list_of_input_to_encode
        self.nb_channels = 0
        self.grid_data_dict = eu.AttrDict()
        self.row, self.col = config_env.grid_shape[-2], config_env.grid_shape[-1]
        self.resolution = 2*config_env.local_map_depth/self.row
        self.local_map_depth = config_env.local_map_depth
        for key in self.list_of_input_to_encode:
            if 'grid' in key :
                self.nb_channels+=1
                self.grid_data_dict[key] = np.zeros((1, self.row, self.col))
            
        # Action Space
        self.action_list = set_action_list()
        self.forward_velocity_limit = [config_env.vmin, config_env.vmax]
        self.angular_velocity_limit = [config_env.wmin, config_env.wmax]

        # Coordinte of goal
        self.distance_to_goal = np.inf
        self.angle_to_goal = 0

        # Orientation data
        self.goal_orientation = None
        self.robot_orientation = None

        # Flags
        self.flag_collision = False

    def compute_observation_from_data(self, 
                                      static_occupancy_grid,
                                      dynamic_social_grid, 
                                      robot_position, 
                                      robot_orientation, 
                                      goal_position, 
                                      linear_velocity=0, 
                                      angular_velocity=0,
                                      goal_yaw=None):
        r"""
            DRL Agent Step.

            Args:
                local_occupancy_grid : numpy.array of size (1, N, N)

                goal_polar_coordinates : numpy.array of size (1, 2)
                
            Returns:
                int x int : linear and angular velocity
        """
        observation = eu.AttrDict()
        
        self.update_grid_dict(
                            static_occupancy_grid=static_occupancy_grid,
                            dynamic_social_grid=dynamic_social_grid
        )

        raycast_map = None
        closest_obs_map = None

        self.goal_orientation = goal_yaw
        self.robot_orientation = robot_orientation
        
        closest_obs_map, distance_to_closest_object, angle_to_closest_object = distance_and_angle_to_nearest_obstacle_from_occupancy_grid(
                                                                                                                        occupancy_grid=self.grid_data_dict.static_occupancy_grid[0,:,:],
                                                                                                                        resolution = self.resolution
                                                                                                                        )

        if distance_to_closest_object < 0.1 :
            self.flag_collision = True
        else :
            self.flag_collision = False

        if 'raycast' in self.list_of_input_to_encode :
            raycast_distance, raycast_angles, raycast_map = raycast_from_occupancy_grid(
                                        occupancy_grid=self.grid_data_dict.static_occupancy_grid[0,:,:],
                                        resolution = self.resolution,
                                        field_of_view=360,
                                        angle_between_rays=10,
                                        draw_ray = True
                    )
            raycast_distance = np.clip(raycast_distance-0.2, 0, self.local_map_depth)

        
        for key in self.list_of_input_to_encode :
            if 'grid' in key:
                observation[key] = self.grid_data_dict[key]

            elif key == 'raycast' :
                observation.raycast = raycast_distance
                
            elif key == 'angle_to_goal_position' :
                self.angle_to_goal = measure_relative_angle(robot_position, robot_orientation, goal_position)
                observation.angle_to_goal_position = self.angle_to_goal
                
            elif key == 'distance_to_goal_position' :
                self.distance_to_goal = measure_distance(robot_position, goal_position)
                observation.distance_to_goal_position = self.distance_to_goal

            elif key == 'angle_to_closest_object'  :
                observation.angle_to_closest_object = angle_to_closest_object

            elif key == 'distance_to_closest_object' :
                observation.distance_to_closest_object = distance_to_closest_object
            
            elif key == 'linear_velocity':
                observation.linear_velocity = linear_velocity
            
            elif key == 'angular_velocity':
                observation.angular_velocity = angular_velocity
            else:
                raise Warning('The input name was not recognized')

        return observation, raycast_map, closest_obs_map



    def predict(self, agent, observation):
        r''' Predict an action from an RL agent and an observation '''
        if self.flag_collision:
            print('Warning !!!!! The robot is close to an obstacle')

        if isinstance(agent, SACBase) :
            # Continuous actions
            prediction = agent.predict_action(observation)

        elif isinstance(agent, DDQN) :
            # Discrete actions
            action_id = agent.predict_action(observation)
            prediction = self.discrete_action_list[action_id]
        else :
            raise ValueError('Problem with agent, it has to be a DDQN or a SACBase')

        if self.distance_to_goal > 0.3:
            action = prediction
        else:
            if np.abs(constraint_angle(self.goal_orientation - self.robot_orientation)) > np.pi/12:
                lin_vel=0
                ang_vel =  0.5*np.sign(constraint_angle(self.goal_orientation - self.robot_orientation))
                action = lin_vel, ang_vel
            else:
                action = (0,0)

        return action, prediction
    


    def update_grid_dict(self, static_occupancy_grid,  dynamic_social_grid):
        ''' Update the grid dict with the static_occupancy_grid and dynamic_social_grid '''
        
        for key in self.grid_data_dict:
            if key == 'static_occupancy_grid':
                if np.shape(static_occupancy_grid)[-1] != self.col :
                    static_occupancy_grid = cv2.resize(static_occupancy_grid, (self.row,self.col), interpolation = cv2.INTER_NEAREST)
                self.grid_data_dict[key][0,:,:] = static_occupancy_grid
            elif key == 'dynamic_social_grid':
                if np.shape(dynamic_social_grid)[-1] != self.col :
                    dynamic_social_grid = cv2.resize(dynamic_social_grid, (self.row,self.col), interpolation = cv2.INTER_NEAREST)
                self.grid_data_dict[key][0,:,:] = dynamic_social_grid

    



