# This file, part of Social MPC in the WP6 of the Spring project,
# is part of a project that has received funding from the 
# European Union’s Horizon 2020 research and innovation programme
#  under grant agreement No 871245.
# 
# Copyright (C) 2020-2022 by Inria
# Authors : Chris Reinke, Alex Auternaud, Timothée Wintz
# chris.reinke@inria.fr
# alex.auternaud@inria.fr
# timothee.wintz@inria.fr

#!/usr/bin/env python3
import numpy as np
import tf
from scipy.interpolate import RectBivariateSpline
import jax.numpy as jnp
from jax import grad, jit, vmap, jacfwd, custom_jvp, partial


import numpy as np

class TemplatedConfig:
    template = {}
    def __init__(self, dict):
        for k in self.template.keys():
            if k not in dict:
                raise ValueError("Missing config key {0}".format(k))
            elif type(self.template[k]) == list:
                if dict[k] not in self.template[k]:
                    raise ValueError("Invalid config key {0} (must be one of {1})".format(k, str(self.template[k])))
            elif type(self.template[k]) == tuple and self.template[k][0] == np.array:
                try:
                    self.__setattr__(k, np.array(dict[k]))
                    assert(self.__getattribute__(k).shape == self.template[k][1])
                except:
                    raise ValueError("Invalid config key {0} (must be numpy array)".format(k))
            elif type(dict[k]) != self.template[k]:
                raise ValueError("Invalid config key {0} (must be {1})".format(k,  self.template[k]))
            else:
                self.__setattr__(k, dict[k])


class RunDemoConfig(TemplatedConfig):
    template = {
        "logging_mode": str,
        "max_steps": int,
        "end_time": int,
        "save_data": bool,
        "h_idx_to_save": int,
        "plot_render": bool,
    }


class ControllerConfig(TemplatedConfig):
    template = {
        "logging_mode": str,
        "h" : float,
        # "fw_horizon" : int,
        "fw_time": float,
        "u_lb": (np.array, (3,)),
        "u_ub": (np.array, (3,)),
        "reg_parameter": (np.array, (3,)),
        "max_acceleration": (np.array, (3,)),
        "max_iter_optim": int,
        "weights_description": list,
        "goto_target_dim": int,
        "human_target_dim": int,
        "pan_target_dim": int,
        "target_pose_goto_dim": int,
        "target_pose_look_dim": int,
        "goto_target_id": int,
        "goto_target_pose": (np.array, (5,)),
        "human_target_id": int,
        "pan_target_id": int,
        "min_dist_to_robot": float,
        "check_object_fw_traj_offset": float,
        "relative_h_pose": (np.array, (2,)),
        "loss": int,
        "goto_weight" : float,
        "human_features_weight": float,
        "object_cost_map_weight": float,
        "social_cost_map_weight": float,
        "wall_avoidance_weight": float,
        "cost_map_weight": float,
        "loss_rad": (np.array, (3,7)),
        "loss_coef": (np.array, (3,7)),
        "group_detection_stride": float,
        "step_meshes": int,
        "waypoint_distance": float,
        "orientation_distance_threshold":float,
        "wall_avoidance_points": (np.array,(1, 2)),
        "path_loop_time": float,
        "goal_loop_time": float,
        "goal_finder_enabled": bool,
        "path_planner_enabled": bool,
        "update_goals_enabled": bool,
        "dilation_iter": int
    }


class RobotConfig(TemplatedConfig):
    template = {
        "base_footprint2base_link" : (np.array, (3,)),
        "base_link2wheel_left_link" : (np.array, (3,)),
        "base_link2wheel_right_link" : (np.array, (3,)),
        "base_link2caster_left_link" : (np.array, (3,)),
        "base_link2caster_right_link" : (np.array, (3,)),
        "base_link2head_1_link" : (np.array, (3,)),
        "base_link2head_2_link" : (np.array, (3,)),
        "base_link2head_front_camera_link" : (np.array, (3,)),
        "base_link2head_front_camera_optical_frame" : (np.array, (3,)),
        "base_footprint_radius": float,
        "wheel_radius" : float,
        "wheel_width" : float,
        "wheel_separation" : float,
        "wheel_torque" : float,
        "wheel_velocity" : float,
        "caster_radius" : float,
        "caster_width" : float,
        "caster_separation_x" : float,
        "caster_offset_y" : float,
        "caster_offset_z" : float,
        "total_mass" : float,
        "wheel_mass" : float,
        "L1" : (np.array, (4,4)),
        "L2" : (np.array, (4,4)),
        "L3" : (np.array, (4,4)),
        "local_map_depth": float,
        "local_map_px_size":float,
        "max_speed_motor_wheel" : float,
        "max_forward_speed": float,
        "max_backward_speed": float,
        "max_forward_acc": float,
        "max_ang_vel": float,
        "max_ang_acc": float,
        "max_pan_angle" : float,
        "min_pan_angle" : float,
        "max_tilt_angle" : float,
        "min_tilt_angle" : float,
        "max_speed_motor_head" : float,
        "head_friction" : float,
        "head_damping" : float,
        "head_cam_fov_h" : float,
        "head_cam_fov_v" : float,
        "base_cam_fov_h" : float,
        "has_pan" : bool,
        "has_tilt" : bool
    }


class RobotState:
    pass


def constraint_angle(angle, min_value=-np.pi, max_value=np.pi):

    length = max_value - min_value

    # if angle > max_value:
    #     diff = angle - max_value
    #     new_angle = min_value + (diff % length)
    # elif angle < min_value:
    #     diff = min_value - angle
    #     new_angle = max_value - (diff % length)
    # else:
    #     new_angle = angle
    new_angle = np.where(angle > max_value, min_value + ((angle - max_value) % length), angle)
    new_angle = np.where(angle < min_value, max_value - ((min_value - angle) % length), new_angle)
    return new_angle


def local_to_global_jax(robot_position, x):
    angle = rotmat_2d_jax(robot_position[:, -1])
    y = vmap(vmpa_dot_jax, in_axes=(0, None))(jnp.moveaxis(angle, -1, 0), x)
    return jnp.array(y + robot_position[:, :2])


def vmpa_dot_jax(a, b):
    return a.dot(b)


def rotmat_2d(angle):
    return np.matrix([[np.cos(angle), -np.sin(angle)],
                      [np.sin(angle), np.cos(angle)]])


def rotmat_2d_jax(angle):
    return jnp.array([[jnp.cos(angle), -jnp.sin(angle)],
                      [jnp.sin(angle), jnp.cos(angle)]])


def local_to_global(robot_position, x):
    angle = robot_position[-1]
    y = rotmat_2d(angle).dot(x) + robot_position[:2]
    return np.array(y).reshape(x.shape)


def global_to_local(robot_position, x):
    angle = robot_position[-1]
    y = rotmat_2d(-angle).dot(x - robot_position[:2])
    return np.array(y).reshape(x.shape)


def pose_stamped_2_mat(p):
    q = p.pose.orientation
    pos = p.pose.position
    T = tf.transformations.quaternion_matrix([q.x,q.y,q.z,q.w])
    T[:3,3] = np.array([pos.x,pos.y,pos.z])
    return T


def get_color(idx):
    idx = idx * 3
    color = ((37 * idx) % 255, (17 * idx) % 255, (29 * idx) % 255)

    return color


def meter_to_px(point, map_size, resolution):
    if point[0] < map_size[0][0]:
        point[0] = map_size[0][0]
    if point[0] > map_size[0][1]:
        point[0] = map_size[0][1]
    if point[1] < map_size[1][0]:
        point[1] = map_size[1][0]
    if point[1] > map_size[1][1]:
        point[1] = map_size[1][1]

    point_px = [0., 0.]
    point_px[0] = abs(map_size[0][0] - point[0])/resolution
    point_px[1] = abs(map_size[1][1] - point[1])/resolution

    return np.rint(point_px)

def optimal_path_2d(travel_time, starting_point, dx, coords, goal=None, max_d_orientation=None, N=100):
    """
    Find the optimal path from starting_point to the zero contour
    of travel_time. dx is the grid spacing
    Solve the equation x_t = - grad t / | grad t |
    """
    
    grad_t_y, grad_t_x = np.gradient(travel_time, dx)
    if isinstance(travel_time, np.ma.MaskedArray):
        grad_t_y[grad_t_y.mask] = 0.0
        grad_t_y = grad_t_y.data
        grad_t_x[grad_t_x.mask] = 0.0
        grad_t_x = grad_t_x.data

    d_interp = RectBivariateSpline(coords[1], coords[0],
                                   travel_time)
    gradx_interp = RectBivariateSpline(coords[1], coords[0],
                                       grad_t_x)
    grady_interp = RectBivariateSpline(coords[1], coords[0],
                                       grad_t_y)

    def get_velocity(position):
        """ return normalized velocity at pos """
        x, y = position
        vel = np.array([gradx_interp(y, x)[0][0],
                        grady_interp(y, x)[0][0]])
        return vel / np.linalg.norm(vel)

    def get_distance(position):
        x, y = position
        return d_interp(y, x)

    def euler_point_update(pos, ds):
        return pos - get_velocity(pos) * ds

    def runge_kutta(pos, ds):
        """ Fourth order Runge Kutta point update """
        k1 = ds * get_velocity(pos)
        k2 = ds * get_velocity(pos - k1/2.0)
        k3 = ds * get_velocity(pos - k2/2.0)
        k4 = ds * get_velocity(pos - k3)
        v = (k1 + 2*k2 + 2*k3 + k4)/6.0
        return pos - v, v

    # def sym(x, v):
    #     x = x[0], 0.5 * (coords[1][-1] + coords[1][0]) - x[1]
    #     v = v[0], -v[1]
    #     return x, v
    #
    # starting_point, _ = sym(starting_point, [0, 0])

    x, v = runge_kutta(starting_point, dx)
    xl, yl, vxl, vyl, dl = [], [], [], [], []
    starting_distance = get_distance(starting_point)

    for i in range(N):
        # xp, vp = sym(x, v)
        d = get_distance(x)[0][0]
        # if negative value d < 0, waypoint in an obstacle
        if d < 0:
            return False
        # goal is implicit defined in the travel_time map but we need to test it explicitly because we can not really trust d
        if ((goal is not None and max_d_orientation is not None) and
                (np.linalg.norm(goal - np.asarray(x)) < max_d_orientation)):
            if dl:
                break
        xl.append(x[0])
        yl.append(x[1])
        vxl.append(v[0])
        vyl.append(v[1])
        dl.append(d)
        x, v = runge_kutta(x, dx)

    return xl, yl, vxl, vyl, dl
