#!/usr/bin/env python3
import numpy as np
import tf
import plotly.graph_objects as go
import cv2
import plotly.express as px
import os
from copy import deepcopy


def set_action_list(linear_velocties_list = [0, 0.1, 0.2, 0.4], angular_velocities_list = [-0.4, -0.2, -0.1, 0, 0.1, 0.2, 0.4]) -> list:
    r""" Function that returns a list of discrete pairs of actions 
    
    Args : 
        linear_velocties_list (list)
            default value : [0, 0.2, 0.4, 0.6]
        
        angular_velocities_list (list)
            default value :  [-np.pi/4, -np.pi/6, -np.pi/12, 0, np.pi/12, np.pi/6, np.pi/4]
    
    Return :
        actions (list)
    
    """
    actions = []
    for linear_velocity in linear_velocties_list :
        for angular_velocity in angular_velocities_list:
            if linear_velocity == 0 and angular_velocity == 0:
                pass
            else :    
                
                # a single step of simulation is making the robot moving forward of 10 cm
                # So with 5 loop of simulation the robot is moving forward with 0.5m
                actions.append([linear_velocity, round(angular_velocity,1)])
            
        
    return(actions)


def measure_distance(current_position, goal_position):
    return np.linalg.norm(np.array(current_position) - np.array(goal_position))

def measure_relative_angle(current_position, current_orientation, goal_position):
    """Measures the relative angle between two objects or points.

    Args:
        current_position : (np.array)

        current_orientation : (float)

        goal_position : (np.array)

    Returns: The angle can be positive or negative indicating the direction.

    Note : function inspired from : https://gitlab.inria.fr/spring/wp6_robot_behavior/multiparty_interaction_simulator/-/blob/main/mpi_sim/utils/measurements.py 
    """


    position_diff = np.array(goal_position) - np.array(current_position)

    if position_diff[0] == 0. and position_diff[1] == 0.:
        angle = 0.
    else:
        angle_between_p1_and_p2 = -np.pi/2 + np.arctan2(position_diff[0], -position_diff[1])
        angle = constraint_angle(angle_between_p1_and_p2 - current_orientation)

    return angle

def global_coord_to_local_coord(position_ari , orientation_ari, entities_positions_global):
    ''' Take the global coordinates (in float) and transform it to local coordinates (in float) '''
    # Translation (ARI is now the reference position)
    entities_positions_local = np.array(entities_positions_global) - np.array(position_ari)
    # Rotation (ARI is now the reference orientation)
    rotation_matrix = np.array([[np.cos(orientation_ari), -np.sin(orientation_ari)],[np.sin(orientation_ari), np.cos(orientation_ari)]])
    entities_positions_local = np.transpose(np.dot(rotation_matrix, np.transpose(entities_positions_local)))
    assert len(entities_positions_local) == len(entities_positions_global)
    return entities_positions_local

def transform_position_to_map_coordinate(position, map_area, map_resolution):
    r"""Transform a position (x, y) in the simulation to a coordinate (x, y) of a map of the simulation.

    Args:
        position (x,y): x,y - Position that should be transformed.
        map_area (tuple): Area ((x_min, x_max), (y_min, y_max)) of the simulation that the map is depicting.
        map_resolution (float): Resolution of the map in m per cells.

    Returns: Map coordinate as a tuple (x, y).
    """
    x_min = map_area[0][0]
    y_min = map_area[1][0]
    x = int((position[0] - x_min) / map_resolution)
    y = int((position[1] - y_min) / map_resolution)
    return x, y



def constraint_angle(angle, min_value=-np.pi, max_value=np.pi):
        """Constrains the given angle between -pi and pi.
        
        Note : function inspired from : https://gitlab.inria.fr/spring/wp6_robot_behavior/multiparty_interaction_simulator/-/blob/main/mpi_sim/utils/measurements.py """

        length = max_value - min_value
        new_angle = np.where(angle > max_value, min_value + ((angle - max_value) % length), angle)
        new_angle = np.where(angle < min_value, max_value - ((min_value - angle) % length), new_angle)

        return new_angle

def pose_stamped_2_mat(p):
    q = p.pose.orientation
    pos = p.pose.position
    T = tf.transformations.quaternion_matrix([q.x,q.y,q.z,q.w])
    T[:3,3] = np.array([pos.x,pos.y,pos.z])
    return T


def get_color(idx):
    idx = idx * 3
    color = ((37 * idx) % 255, (17 * idx) % 255, (29 * idx) % 255)

    return color


def meter_to_px(point, map_size, resolution):
    if point[0] < map_size[0][0]:
        point[0] = map_size[0][0]
    if point[0] > map_size[0][1]:
        point[0] = map_size[0][1]
    if point[1] < map_size[1][0]:
        point[1] = map_size[1][0]
    if point[1] > map_size[1][1]:
        point[1] = map_size[1][1]

    point_px = [0., 0.]
    point_px[0] = abs(map_size[0][0] - point[0])/resolution
    point_px[1] = abs(map_size[1][1] - point[1])/resolution

    return np.rint(point_px)

def discrete_colorscale(bvals, colors):
    """
    bvals - list of values bounding intervals/ranges of interest
    colors - list of rgb or hex colorcodes for values in [bvals[k], bvals[k+1]],0<=k < len(bvals)-1
    returns the plotly  discrete colorscale

    From : https://chart-studio.plotly.com/~empet/15229/heatmap-with-a-discrete-colorscale/#/
    """

    dcolorscale = [] #discrete colorscale
    dcolorscale.extend([[bvals[0], colors[0]], [bvals[1], colors[1]]])
    return dcolorscale    

def rotmat_2d(angle):
    return np.matrix([[np.cos(angle), -np.sin(angle)],
                      [np.sin(angle), np.cos(angle)]])

        
def local_to_global(robot_position, x):
    angle = robot_position[-1]
    y = rotmat_2d(angle).dot(x) + robot_position[:2]
    return np.array(y).reshape(x.shape)
            
            
def plot_img(img, name = 'img'):
    # bvals = [0, 1]
    # colors = ['#ffffff', '#000000']
    # dcolorsc = discrete_colorscale(bvals, colors)
    # heatmap = go.Heatmap(z=self.local_map_static.T,
    #                     colorscale = dcolorsc,
    #                     zmin=0, zmax=1,
    #                     showscale=False 
    #                     )
    # fig = go.Figure(data = heatmap, layout = go.Layout(height =  800, width =  800))
    img = deepcopy(img)
    if not os.path.exists('/home/ros/robot_behavior_ws/local_maps'):
            os.makedirs('/home/ros/robot_behavior_ws/local_maps')
    fig = px.imshow(np.flip(np.flip(img.T, axis = 0), axis = 1))
    fig.write_image('/home/ros/robot_behavior_ws/local_maps/{}.png'.format(name))



        
        
        

